# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#
# Regression test class builtins
#

import functools
import reframe.core.parameters as parameters
import reframe.core.variables as variables
import reframe.core.fixtures as fixtures
import reframe.core.hooks as hooks
import reframe.utility as utils
from reframe.core.deferrable import deferrable, _DeferredPerformanceExpression


__all__ = ['deferrable', 'deprecate', 'final', 'fixture', 'loggable',
           'loggable_as', 'parameter', 'performance_function', 'required',
           'require_deps', 'run_before', 'run_after', 'sanity_function',
           'variable']

parameter = parameters.TestParam
variable = variables.TestVar
required = variables.Undefined
deprecate = variables.TestVar.create_deprecated
fixture = fixtures.TestFixture


def final(fn):
    '''Indicate that a function is final and cannot be overridden.'''

    fn._rfm_final = True
    return fn


# Hook-related builtins

def run_before(stage):
    '''Attach the decorated function before a certain pipeline stage.

    The function will run just before the specified pipeline stage and it
    cannot accept any arguments except ``self``. This decorator can be
    stacked, in which case the function will be attached to multiple pipeline
    stages. See above for the valid ``stage`` argument values.

    :param stage: The pipeline stage where this function will be attached to.
        See :ref:`pipeline-hooks` for the list of valid stage values.
    '''
    return hooks.attach_to('pre_' + stage)


def run_after(stage):
    '''Attach the decorated function after a certain pipeline stage.

    This is analogous to :func:`~RegressionMixin.run_before`, except that the
    hook will execute right after the stage it was attached to. This decorator
    also supports ``'init'`` as a valid ``stage`` argument, where in this
    case, the hook will execute right after the test is initialized (i.e.
    after the :func:`__init__` method is called) and before entering the
    test's pipeline. In essence, a post-init hook is equivalent to defining
    additional :func:`__init__` functions in the test. The following code

    .. code-block:: python

       class MyTest(rfm.RegressionTest):
           @run_after('init')
           def foo(self):
               self.x = 1

    is equivalent to

    .. code-block:: python

       class MyTest(rfm.RegressionTest):
           def __init__(self):
               self.x = 1

    .. versionchanged:: 3.5.2
       Add support for post-init hooks.

    '''
    return hooks.attach_to('post_' + stage)


require_deps = hooks.require_deps


# Sanity and performance function builtins

def sanity_function(fn):
    '''Decorate a test member function to mark it as a sanity check.

    This decorator will convert the given function into a
    :func:`~RegressionMixin.deferrable` and mark it to be executed during the
    test's sanity stage. When this decorator is used, manually assigning a
    value to :attr:`~RegressionTest.sanity_patterns` in the test is not
    allowed.

    Decorated functions may be overridden by derived classes, and derived
    classes may also decorate a different method as the test's sanity
    function. Decorating multiple member functions in the same class is not
    allowed. However, a :class:`RegressionTest` may inherit from multiple
    :class:`RegressionMixin` classes with their own sanity functions. In this
    case, the derived class will follow Python's `MRO
    <https://docs.python.org/3/library/stdtypes.html#class.__mro__>`_ to find
    a suitable sanity function.

    .. versionadded:: 3.7.0
    '''

    _def_fn = deferrable(fn)
    setattr(_def_fn, '_rfm_sanity_fn', True)
    return _def_fn


def performance_function(unit, *, perf_key=None):
    '''Decorate a test member function to mark it as a performance metric
    function.

    This decorator converts the decorated method into a performance deferrable
    function (see ":ref:`deferrable-performance-functions`" for more details)
    whose evaluation is deferred to the performance stage of the regression
    test. The decorated function must take a single argument without a default
    value (i.e. ``self``) and any number of arguments with default values. A
    test may decorate multiple member functions as performance functions,
    where each of the decorated functions must be provided with the unit of
    the performance quantity to be extracted from the test. Any performance
    function may be overridden in a derived class and multiple bases may
    define their own performance functions. In the event of a name conflict,
    the derived class will follow Python's `MRO
    <https://docs.python.org/3/library/stdtypes.html#class.__mro__>`_ to
    choose the appropriate performance function. However, defining more than
    one performance function with the same name in the same class is
    disallowed.

    The full set of performance functions of a regression test is stored under
    :attr:`~reframe.core.pipeline.RegressionTest.perf_variables` as key-value
    pairs, where, by default, the key is the name of the decorated member
    function, and the value is the deferred performance function itself.
    Optionally, the key under which a performance function is stored in
    :attr:`~reframe.core.pipeline.RegressionTest.perf_variables` can be
    customised by passing the desired key as the ``perf_key`` argument to this
    decorator.

    :param unit: A string representing the measurement unit of this metric.

    .. versionadded:: 3.8.0

    '''

    if not isinstance(unit, str):
        raise TypeError('performance unit must be a string')

    if perf_key and not isinstance(perf_key, str):
        raise TypeError("'perf_key' must be a string")

    def _deco_wrapper(func):
        if not utils.is_trivially_callable(func, non_def_args=1):
            raise TypeError(
                f'performance function {func.__name__!r} has more '
                f'than one argument without a default value'
            )

        @functools.wraps(func)
        def _perf_fn(*args, **kwargs):
            return _DeferredPerformanceExpression(
                func, unit, *args, **kwargs
            )

        _perf_key = perf_key if perf_key else func.__name__
        setattr(_perf_fn, '_rfm_perf_key', _perf_key)
        return _perf_fn

    return _deco_wrapper


def loggable_as(name):
    '''Mark a property as loggable.

    :param name: An alternative name that will be used for logging
        this property. If :obj:`None`, the name of the decorated
        property will be used.
    :raises ValueError: if the decorated function is not a property.

    .. versionadded:: 3.10.2

    :meta private:

    '''
    def _loggable(fn):
        if not hasattr(fn, 'fget'):
            raise ValueError('decorated function does not '
                             'look like a property')

        # Mark property as loggable
        #
        # NOTE: Attributes cannot be set on property objects, so we
        # set the attribute on one of its functions
        prop_name = fn.fget.__name__
        fn.fget._rfm_loggable = (prop_name, name)
        return fn

    return _loggable


loggable = loggable_as(None)
loggable.__doc__ = '''Equivalent to :func:`loggable_as(None) <loggable_as>`.'''



# Copyright 2016-2020 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import os
import sys
import yaml

import reframe.core.exceptions as errors
import reframe.core.runtime as runtime


def _emit_gitlab_pipeline(testcases, child_pipeline_opts):
    config = runtime.runtime().site_config

    # Collect the necessary ReFrame invariants
    program = 'reframe'
    prefix = 'rfm-stage/${CI_COMMIT_SHORT_SHA}'
    checkpath = config.get('general/0/check_search_path')
    recurse = config.get('general/0/check_search_recursive')
    verbosity = 'v' * config.get('general/0/verbose')

    def rfm_command(testcase):
        # Ignore the first argument, it should be '<builtin>'
        config_opt = ' '.join([f'-C {arg}' for arg in config.sources[1:]])

        report_file = f'{testcase.check.unique_name}-report.json'
        if testcase.level:
            restore_files = ','.join(
                f'{t.check.unique_name}-report.json' for t in tc.deps
            )
        else:
            restore_files = None

        return ' '.join([
            program,
            f'--prefix={prefix}', config_opt,
            f'{" ".join("-c " + c for c in checkpath)}',
            f'-R' if recurse else '',
            f'--report-file={report_file}',
            f'--restore-session={restore_files}' if restore_files else '',
            f'--report-junit={testcase.check.unique_name}-report.xml',
            f'{"".join("-" + verbosity)}' if verbosity else '',
            '-n', f"'^{testcase.check.unique_name}$'", '-r',
            *child_pipeline_opts
        ])

    max_level = 0   # We need the maximum level to generate the stages section
    json = {
        'cache': {
            'key': '${CI_COMMIT_REF_SLUG}',
            'paths': ['rfm-stage/${CI_COMMIT_SHORT_SHA}']
        },
        'stages': []
    }

    # Name of the image used for CI. If user does not explicitly provide
    # image keyword on the top of CI script, this variable does not exist
    image_name = os.getenv('CI_JOB_IMAGE')
    if image_name:
        json['image'] = image_name

    for tc in testcases:
        json[f'{tc.check.unique_name}'] = {
            'stage': f'rfm-stage-{tc.level}',
            'script': [rfm_command(tc)],
            'artifacts': {
                'paths': [f'{tc.check.unique_name}-report.json']
            },
            'needs': [t.check.unique_name for t in tc.deps]
        }
        max_level = max(max_level, tc.level)

    json['stages'] = [f'rfm-stage-{m}' for m in range(max_level+1)]
    return json


def emit_pipeline(fp, testcases, child_pipeline_opts=None, backend='gitlab'):
    if backend != 'gitlab':
        raise errors.ReframeError(f'unknown CI backend {backend!r}')

    child_pipeline_opts = child_pipeline_opts or []
    yaml.dump(_emit_gitlab_pipeline(testcases, child_pipeline_opts), stream=fp,
              indent=2, sort_keys=False, width=sys.maxsize)



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import re
import json

import reframe.utility as util
import reframe.utility.jsonext as jsonext
from reframe.core.backends import (getlauncher, getscheduler)
from reframe.core.environments import (Environment, ProgEnvironment)
from reframe.core.exceptions import ConfigError
from reframe.core.logging import getlogger
from reframe.core.modules import ModulesSystem


class _ReadOnlyInfo:
    __slots__ = ('_info',)
    _known_attrs = ()

    def __init__(self, info):
        self._info = info

    def __deepcopy__(self, memo):
        # This is a read-only object; simply return ourself
        return self

    def __getattr__(self, name):
        if name in self._known_attrs:
            return self._info.get(name, None)
        else:
            raise AttributeError(
                f'{type(self).__qualname__!r} object has no attribute {name!r}'
            )

    def __setattr__(self, name, value):
        if name in self._known_attrs:
            raise AttributeError(f'attribute {name!r} is not writeable')
        else:
            super().__setattr__(name, value)


class ProcessorInfo(_ReadOnlyInfo, jsonext.JSONSerializable):
    '''A representation of a processor inside ReFrame.

    You can access all the keys of the `processor configuration object
    <config_reference.html#processor-info>`__.

    .. versionadded:: 3.5.0

    .. warning::
       Users may not create :class:`ProcessorInfo` objects directly.

    '''

    __slots__ = ()
    _known_attrs = (
        'arch', 'num_cpus', 'num_cpus_per_core',
        'num_cpus_per_socket', 'num_sockets', 'topology'
    )

    @property
    def info(self):
        '''All the available information from the configuration.

        :type: :class:`dict`
        '''
        return self._info

    @property
    def num_cores(self):
        '''Total number of cores.

        :type: integral or :class:`None`
        '''
        if self.num_cpus and self.num_cpus_per_core:
            return self.num_cpus // self.num_cpus_per_core
        else:
            return None

    @property
    def num_cores_per_socket(self):
        '''Number of cores per socket.

        :type: integral or :class:`None`
        '''
        if self.num_cores and self.num_sockets:
            return self.num_cores // self.num_sockets
        else:
            return None

    @property
    def num_numa_nodes(self):
        '''Number of NUMA nodes.

        :type: integral or :class:`None`
        '''
        if self.topology and 'numa_nodes' in self.topology:
            return len(self.topology['numa_nodes'])
        else:
            return None

    @property
    def num_cores_per_numa_node(self):
        '''Number of cores per NUMA node.

        :type: integral or :class:`None`
        '''

        if self.num_numa_nodes and self.num_cores:
            return self.num_cores // self.num_numa_nodes
        else:
            return None


class DeviceInfo(_ReadOnlyInfo, jsonext.JSONSerializable):
    '''A representation of a device inside ReFrame.

    You can access all the keys of the `device configuration object
    <config_reference.html#device-info>`__.

    .. versionadded:: 3.5.0

    .. warning::
       Users may not create :class:`DeviceInfo` objects directly.

    '''

    __slots__ = ()
    _known_attrs = ('type', 'arch')

    @property
    def info(self):
        '''All the available information from the configuration.

        :type: :class:`dict`
        '''
        return self._info

    @property
    def num_devices(self):
        '''Number of devices of this type.

        It will return 1 if it wasn't set in the configuration.

        :type: integral
        '''
        return self._info.get('num_devices', 1)

    @property
    def device_type(self):
        '''The type of the device.

        :type: :class:`str` or :class:`None`
        '''
        return self.type


class SystemPartition(jsonext.JSONSerializable):
    '''A representation of a system partition inside ReFrame.

    .. warning::
       Users may not create :class:`SystemPartition` objects directly.
    '''

    def __init__(self, *, parent, name, sched_type, launcher_type,
                 descr, access, container_runtime, container_environs,
                 resources, local_env, environs, max_jobs, prepare_cmds,
                 processor, devices, extras, features, time_limit):
        getlogger().debug(f'Initializing system partition {name!r}')
        self._parent_system = parent
        self._name = name
        self._sched_type = sched_type
        self._scheduler = None
        self._launcher_type = launcher_type
        self._descr = descr
        self._access = access
        self._container_runtime = container_runtime
        self._container_environs = container_environs
        self._local_env = local_env
        self._environs = environs
        self._max_jobs = max_jobs
        self._prepare_cmds = prepare_cmds
        self._resources = {r['name']: r['options'] for r in resources}
        self._processor = ProcessorInfo(processor)
        self._devices = [DeviceInfo(d) for d in devices]
        self._extras = extras
        self._features = features
        self._time_limit = time_limit

    @property
    def access(self):
        '''The scheduler options for accessing this system partition.

        :type: :class:`List[str]`
        '''
        return util.SequenceView(self._access)

    @property
    def descr(self):
        '''The description of this partition.

        :type: :class:`str`
        '''
        return self._descr

    @property
    def environs(self):
        '''The programming environments associated with this system partition.

        :type: :class:`List[ProgEnvironment]`
        '''

        return util.SequenceView(self._environs)

    @property
    def container_runtime(self):
        '''The default container runtime of this partition.

        :type: :class:`str` or ``None``
        '''
        return self._container_runtime

    @property
    def container_environs(self):
        '''Environments associated with the different container platforms.

        :type: :class:`Dict[str, Environment]`
        '''

        return util.MappingView(self._container_environs)

    @property
    def fullname(self):
        '''Return the fully-qualified name of this partition.

        The fully-qualified name is of the form
        ``<parent-system-name>:<partition-name>``.

        :type: :class:`str`
        '''
        return f'{self._parent_system}:{self._name}'

    @property
    def local_env(self):
        '''The local environment associated with this partition.

        :type: :class:`Environment`
        '''
        return self._local_env

    @property
    def max_jobs(self):
        '''The maximum number of concurrent jobs allowed on this partition.

        :type: integral
        '''
        return self._max_jobs

    @property
    def time_limit(self):
        '''The time limit that will be used when submitting jobs to this
        partition.

        :type: :class:`str` or :obj:`None`

        .. versionadded:: 3.11.0
        '''
        return self._time_limit

    @property
    def prepare_cmds(self):
        '''Commands to be emitted before loading the modules.

        :type: :class:`List[str]`
        '''
        return self._prepare_cmds

    @property
    def name(self):
        '''The name of this partition.

        :type: :class:`str`
        '''
        return self._name

    @property
    def resources(self):
        '''The resources template strings associated with this partition.

        This is a dictionary, where the key is the name of a resource and the
        value is the scheduler options or directives associated with this
        resource.

        :type: :class:`Dict[str, List[str]]`

        '''

        return util.MappingView(self._resources)

    @property
    def scheduler(self):
        '''The backend scheduler of this partition.

        :type: :class:`reframe.core.schedulers.JobScheduler`.

        .. note::
           .. versionchanged:: 2.8
              Prior versions returned a string representing the scheduler and
              job launcher combination.
           .. versionchanged:: 3.2
              The property now stores a :class:`JobScheduler` instance.

        '''
        if self._scheduler is None:
            self._scheduler = self._sched_type(part_name=self.name)

        return self._scheduler

    @property
    def launcher_type(self):
        '''The type of the backend launcher of this partition.

        .. versionadded:: 3.2

        :type: a subclass of :class:`reframe.core.launchers.JobLauncher`.
        '''
        return self._launcher_type

    def get_resource(self, name, **values):
        '''Instantiate managed resource ``name`` with ``value``.

        :meta private:
        '''

        ret = []
        for r in self._resources.get(name, []):
            try:
                ret.append(r.format(**values))
            except KeyError:
                pass

        return ret

    def environment(self, name):
        '''Return the partition environment named ``name``.'''

        for e in self.environs:
            if e.name == name:
                return e

        return None

    @property
    def processor(self):
        '''Processor information for the current partition.

        .. versionadded:: 3.5.0

        :type: :class:`reframe.core.systems.ProcessorInfo`
        '''
        return self._processor

    @property
    def devices(self):
        '''A list of devices in the current partition.

        .. versionadded:: 3.5.0

        :type: :class:`List[reframe.core.systems.DeviceInfo]`
        '''
        return self._devices

    @property
    def extras(self):
        '''User defined properties associated with this partition.

        These extras are defined in the configuration.

        .. versionadded:: 3.5.0

        :type: :class:`Dict[str, object]`
        '''
        return self._extras

    @property
    def features(self):
        '''User defined features associated with this partition.

        These features are defined in the configuration.

        .. versionadded:: 3.11.0

        :type: :class:`List[str]`
        '''
        return self._features

    def select_devices(self, devtype):
        '''Return all devices of the requested type:

        :arg devtype: The type of the device info objects to return.
        :returns: A list of :class:`DeviceInfo` objects of the specified type.
        '''
        return [d for d in self.devices if d.device_type == devtype]

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented

        return (self._name == other.name and
                self._sched_type == other._sched_type and
                self._launcher_type == other._launcher_type and
                self._access == other._access and
                self._environs  == other._environs and
                self._resources == other._resources and
                self._local_env == other._local_env)

    def __hash__(self):
        return hash(self.fullname)

    def json(self):
        '''Return a JSON object representing this system partition.'''

        return {
            'name': self._name,
            'descr': self._descr,
            'scheduler': self._sched_type.registered_name,
            'launcher': self._launcher_type.registered_name,
            'access': self._access,
            'container_platforms': [
                {
                    'type': ctype,
                    'modules': [m for m in cpenv.modules],
                    'env_vars': [[n, v] for n, v in cpenv.env_vars.items()]
                }
                for ctype, cpenv in self._container_environs.items()
            ],
            'modules': [m for m in self._local_env.modules],
            'env_vars': [[n, v] for n, v in self._local_env.env_vars.items()],
            'environs': [e.name for e in self._environs],
            'max_jobs': self._max_jobs,
            'resources': [
                {
                    'name': name,
                    'options': options
                }
                for name, options in self._resources.items()
            ]
        }

    def __str__(self):
        return json.dumps(self.json(), indent=2)


class System(jsonext.JSONSerializable):
    '''A representation of a system inside ReFrame.

    .. warning::
       Users may not create :class:`System` objects directly.
    '''

    def __init__(self, name, descr, hostnames, modules_system,
                 preload_env, prefix, outputdir,
                 resourcesdir, stagedir, partitions):
        getlogger().debug(f'Initializing system {name!r}')
        self._name = name
        self._descr = descr
        self._hostnames = hostnames
        self._modules_system = ModulesSystem.create(modules_system)
        self._preload_env = preload_env
        self._prefix = prefix
        self._outputdir = outputdir
        self._resourcesdir = resourcesdir
        self._stagedir = stagedir
        self._partitions = partitions

    @classmethod
    def create(cls, site_config):
        # Create the whole system hierarchy from bottom up
        sysname = site_config.get('systems/0/name')
        partitions = []
        config_save = site_config.subconfig_system

        for p in site_config.get('systems/0/partitions'):
            site_config.select_subconfig(f'{sysname}:{p["name"]}')
            partid = f"systems/0/partitions/@{p['name']}"
            part_name = site_config.get(f'{partid}/name')
            try:
                part_sched = getscheduler(
                    site_config.get(f'{partid}/scheduler')
                )
                part_launcher = getlauncher(
                    site_config.get(f'{partid}/launcher')
                )
            except ConfigError as err:
                # Re-raise with more information
                sys_name = site_config.get('systems/0/name')
                part_fullname = f'{sys_name}:{part_name}'
                raise ConfigError(
                    f'failed to initialize partition {part_fullname!r}'
                ) from err

            part_container_environs = {}
            part_container_runtime = None
            container_platforms = site_config.get(
                f'{partid}/container_platforms')
            for i, p in enumerate(container_platforms):
                ctype = p['type']
                part_container_environs[ctype] = Environment(
                    name=f'__rfm_env_{ctype}',
                    modules=site_config.get(
                        f'{partid}/container_platforms/{i}/modules'
                    ),
                    env_vars=site_config.get(
                        f'{partid}/container_platforms/{i}/env_vars'
                    )
                )
                if p.get('default', None):
                    part_container_runtime = ctype

            if not part_container_runtime and container_platforms:
                # No default set, pick the first one
                part_container_runtime = container_platforms[0]['type']

            env_patt = site_config.get('general/0/valid_env_names') or [r'.*']
            part_environs = [
                ProgEnvironment(
                    name=e,
                    modules=site_config.get(f'environments/@{e}/modules'),
                    env_vars=site_config.get(f'environments/@{e}/env_vars'),
                    extras=site_config.get(f'environments/@{e}/extras'),
                    features=site_config.get(f'environments/@{e}/features'),
                    cc=site_config.get(f'environments/@{e}/cc'),
                    cxx=site_config.get(f'environments/@{e}/cxx'),
                    ftn=site_config.get(f'environments/@{e}/ftn'),
                    cppflags=site_config.get(f'environments/@{e}/cppflags'),
                    cflags=site_config.get(f'environments/@{e}/cflags'),
                    cxxflags=site_config.get(f'environments/@{e}/cxxflags'),
                    fflags=site_config.get(f'environments/@{e}/fflags'),
                    ldflags=site_config.get(f'environments/@{e}/ldflags')
                ) for e in site_config.get(f'{partid}/environs')
                if any(re.match(pattern, e) for pattern in env_patt)
            ]
            partitions.append(
                SystemPartition(
                    parent=site_config.get('systems/0/name'),
                    name=part_name,
                    sched_type=part_sched,
                    launcher_type=part_launcher,
                    descr=site_config.get(f'{partid}/descr'),
                    access=site_config.get(f'{partid}/access'),
                    resources=site_config.get(f'{partid}/resources'),
                    environs=part_environs,
                    container_runtime=part_container_runtime,
                    container_environs=part_container_environs,
                    local_env=Environment(
                        name=f'__rfm_env_{part_name}',
                        modules=site_config.get(f'{partid}/modules'),
                        env_vars=site_config.get(f'{partid}/env_vars')
                    ),
                    max_jobs=site_config.get(f'{partid}/max_jobs'),
                    prepare_cmds=site_config.get(f'{partid}/prepare_cmds'),
                    processor=site_config.get(f'{partid}/processor'),
                    devices=site_config.get(f'{partid}/devices'),
                    extras=site_config.get(f'{partid}/extras'),
                    features=site_config.get(f'{partid}/features'),
                    time_limit=site_config.get(f'{partid}/time_limit')
                )
            )

        # Restore configuration, but ignore unresolved sections or
        # configuration parameters at the system level; if we came up to this
        # point, then all is good at the partition level, which is enough.
        site_config.select_subconfig(config_save, ignore_resolve_errors=True)
        return System(
            name=sysname,
            descr=site_config.get('systems/0/descr'),
            hostnames=site_config.get('systems/0/hostnames'),
            modules_system=site_config.get('systems/0/modules_system'),
            preload_env=Environment(
                name=f'__rfm_env_{sysname}',
                modules=site_config.get('systems/0/modules'),
                env_vars=site_config.get('systems/0/env_vars')
            ),
            prefix=site_config.get('systems/0/prefix'),
            outputdir=site_config.get('systems/0/outputdir'),
            resourcesdir=site_config.get('systems/0/resourcesdir'),
            stagedir=site_config.get('systems/0/stagedir'),
            partitions=partitions
        )

    @property
    def name(self):
        '''The name of this system.

        :type: :class:`str`
        '''
        return self._name

    @property
    def descr(self):
        '''The description of this system.

        :type: :class:`str`
        '''
        return self._descr

    @property
    def hostnames(self):
        '''The hostname patterns associated with this system.

        :type: :class:`List[str]`
        '''
        return self._hostnames

    @property
    def modules_system(self):
        '''The modules system name associated with this system.

        :type: :class:`reframe.core.modules.ModulesSystem`
        '''
        return self._modules_system

    @property
    def preload_environ(self):
        '''The environment to load whenever ReFrame runs on this system.

        .. versionadded:: 2.19

        :type: :class:`reframe.core.environments.Environment`
        '''
        return self._preload_env

    @property
    def prefix(self):
        '''The ReFrame prefix associated with this system.

        :type: :class:`str`
        '''
        return self._prefix

    @property
    def stagedir(self):
        '''The ReFrame stage directory prefix associated with this system.

        :type: :class:`str`
        '''
        return self._stagedir

    @property
    def outputdir(self):
        '''The ReFrame output directory prefix associated with this system.

        :type: :class:`str`
        '''
        return self._outputdir

    @property
    def resourcesdir(self):
        '''Global resources directory for this system.

        This directory may be used for storing large files related to
        regression tests. The value of this directory is controlled by the
        `resourcesdir <config_reference.html#.systems[].resourcesdir>`__
        configuration parameter.

        :type: :class:`str`

        '''
        return self._resourcesdir

    @property
    def partitions(self):
        '''The system partitions associated with this system.

        :type: :class:`List[SystemPartition]`
        '''
        return util.SequenceView(self._partitions)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented

        return (self._name == other._name and
                self._hostnames  == other._hostnames and
                self._partitions == other._partitions)

    def json(self):
        '''Return a JSON object representing this system.'''

        return {
            'name': self._name,
            'descr': self._descr,
            'hostnames': self._hostnames,
            'modules_system': self._modules_system.name,
            'modules': [m for m in self._preload_env.modules],
            'env_vars': [
                [name, value]
                for name, value in self._preload_env.env_vars.items()
            ],
            'prefix': self._prefix,
            'outputdir': self._outputdir,
            'stagedir': self._stagedir,
            'resourcesdir': self._resourcesdir,
            'partitions': [p.json() for p in self._partitions]
        }

    def __str__(self):
        return json.dumps(self.json(), indent=2)

    def __repr__(self):
        return (
            f'{type(self).__name__}( '
            f'name={self._name!r}, descr={self._descr!r}, '
            f'hostnames={self._hostnames!r}, '
            f'modules_system={self.modules_system.name!r}, '
            f'preload_env={self._preload_env!r}, prefix={self._prefix!r}, '
            f'outputdir={self._outputdir!r}, '
            f'resourcesdir={self._resourcesdir!r}, '
            f'stagedir={self._stagedir!r}, partitions={self._partitions!r})'
        )



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#
# SGE backend
#
# - Initial version submitted by Mosè Giordano, UCL (based on the PBS backend)
#

import functools
import re
import time
import xml.etree.ElementTree as ET

import reframe.utility.osext as osext
from reframe.core.backends import register_scheduler
from reframe.core.exceptions import JobSchedulerError
from reframe.core.schedulers.pbs import PbsJobScheduler
from reframe.utility import seconds_to_hms

_run_strict = functools.partial(osext.run_command, check=True)


@register_scheduler('sge')
class SgeJobScheduler(PbsJobScheduler):
    def __init__(self):
        self._prefix = '#$'
        self._submit_timeout = self.get_option('job_submit_timeout')

    def emit_preamble(self, job):
        preamble = [
            self._format_option(f'-N "{job.name}"'),
            self._format_option(f'-o {job.stdout}'),
            self._format_option(f'-e {job.stderr}'),
            self._format_option(f'-wd {job.workdir}')
        ]

        if job.time_limit is not None:
            h, m, s = seconds_to_hms(job.time_limit)
            preamble.append(
                self._format_option(f'-l h_rt=%d:%d:%d' % (h, m, s))
            )

        # Emit the rest of the options
        options = job.options + job.cli_options
        for opt in options:
            if opt.startswith('#'):
                preamble.append(opt)
            else:
                preamble.append(self._format_option(opt))

        return preamble

    def submit(self, job):
        # `-o` and `-e` options are only recognized in command line by the PBS,
        # SGE, and Slurm wrappers.
        cmd = f'qsub -o {job.stdout} -e {job.stderr} {job.script_filename}'
        completed = _run_strict(cmd, timeout=self._submit_timeout)
        jobid_match = re.search(r'^Your job (?P<jobid>\S+)', completed.stdout)
        if not jobid_match:
            raise JobSchedulerError('could not retrieve the job id '
                                    'of the submitted job')

        job._jobid = jobid_match.group('jobid')
        job._submit_time = time.time()

    def poll(self, *jobs):
        if jobs:
            # Filter out non-jobs
            jobs = [job for job in jobs if job is not None]

        if not jobs:
            return

        user = osext.osuser()
        completed = osext.run_command(f'qstat -xml -u {user}')
        if completed.returncode != 0:
            raise JobSchedulerError(
                f'qstat failed with exit code {completed.returncode} '
                f'(standard error follows):\n{completed.stderr}'
            )

        # Index the jobs to poll on their jobid
        jobs_to_poll = {job.jobid: job for job in jobs}

        # Parse the XML
        root = ET.fromstring(completed.stdout)

        # We are iterating over the returned XML and update the status of the
        # jobs relevant to ReFrame; the naming convention of variables matches
        # that of SGE's XML output

        known_jobs = set()  # jobs known to the SGE scheduler
        for queue_info in root:
            # Reads the XML and prints jobs with status belonging to user.
            if queue_info is None:
                raise JobSchedulerError('could not retrieve queue information')

            for job_list in queue_info:
                if job_list.find("JB_owner").text != user:
                    # Not a job of this user.
                    continue

                jobid = job_list.find("JB_job_number").text
                if jobid not in jobs_to_poll:
                    # Not a reframe job
                    continue

                state = job_list.find("state").text
                job = jobs_to_poll[jobid]
                known_jobs.add(job)

                # For the list of known statuses see `man 5 sge_status`
                # (https://arc.liv.ac.uk/SGE/htmlman/htmlman5/sge_status.html)
                if state in ['r', 'hr', 't', 'Rr', 'Rt']:
                    job._state = 'RUNNING'
                elif state in ['qw', 'Rq', 'hqw', 'hRwq']:
                    job._state = 'PENDING'
                elif state in ['s', 'ts', 'S', 'tS', 'T', 'tT', 'Rs',
                               'Rts', 'RS', 'RtS', 'RT', 'RtT']:
                    job._state = 'SUSPENDED'
                elif state in ['Eqw', 'Ehqw', 'EhRqw']:
                    job._state = 'ERROR'
                elif state in ['dr', 'dt', 'dRr', 'dRt', 'ds',
                               'dS', 'dT', 'dRs', 'dRS', 'dRT']:
                    job._state = 'DELETING'
                elif state == 'z':
                    job._state = 'COMPLETED'

        # Mark any "unknown" job as completed
        unknown_jobs = set(jobs) - known_jobs
        for job in unknown_jobs:
            self.log(f'Job {job.jobid} not known to scheduler, '
                     f'assuming job completed')
            job._state = 'COMPLETED'

    def finished(self, job):
        if job.exception:
            raise job.exception

        return job.state == 'COMPLETED'



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#
# Meta-class for creating regression tests.
#

import functools
import types
import collections

import reframe.core.builtins as builtins
import reframe.core.namespaces as namespaces
import reframe.core.parameters as parameters
import reframe.core.variables as variables
import reframe.core.fixtures as fixtures
import reframe.core.hooks as hooks
import reframe.utility as utils

from reframe.core.exceptions import ReframeSyntaxError


class RegressionTestMeta(type):
    class MetaNamespace(namespaces.LocalNamespace):
        '''Custom namespace to control the cls attribute assignment.

        Regular Python class attributes can be overridden by either
        parameters or variables respecting the order of execution.
        A variable or a parameter may not be declared more than once in the
        same class body. Overriding a variable with a parameter or the other
        way around has an undefined behavior. A variable's value may be
        updated multiple times within the same class body. A parameter's
        value cannot be updated more than once within the same class body.
        '''

        def __setitem__(self, key, value):
            if isinstance(value, variables.TestVar):
                # Insert the attribute in the variable namespace
                try:
                    self['_rfm_local_var_space'][key] = value
                    value.__set_name__(self, key)
                except KeyError:
                    raise ReframeSyntaxError(
                        f'variable {key!r} is already declared'
                    ) from None

                # Override the regular class attribute (if present) and return
                self._namespace.pop(key, None)
                return
            elif isinstance(value, parameters.TestParam):
                # Insert the attribute in the parameter namespace
                try:
                    self['_rfm_local_param_space'][key] = value
                except KeyError:
                    raise ReframeSyntaxError(
                        f'parameter {key!r} is already declared in this class'
                    ) from None

                # Override the regular class attribute (if present) and return
                self._namespace.pop(key, None)
                return
            elif isinstance(value, fixtures.TestFixture):
                # Insert the attribute in the fixture namespace
                self['_rfm_local_fixture_space'][key] = value

                # Override the regular class attribute (if present)
                self._namespace.pop(key, None)
                return
            elif key in self['_rfm_local_param_space']:
                raise ReframeSyntaxError(
                    f'cannot redefine parameter {key!r}'
                )
            elif key in self['_rfm_local_fixture_space']:
                raise ReframeSyntaxError(
                    f'cannot redefine fixture {key!r}'
                )
            else:
                # Insert the items manually to overide the namespace clash
                # check from the base namespace.
                self._namespace[key] = value

            # Register functions decorated with either @sanity_function or
            # @performance_variables or @performance_function decorators.
            if hasattr(value, '_rfm_sanity_fn'):
                try:
                    super().__setitem__('_rfm_sanity', value)
                except KeyError:
                    raise ReframeSyntaxError(
                        'the @sanity_function decorator can only be used '
                        'once in the class body'
                    ) from None
            elif hasattr(value, '_rfm_perf_key'):
                try:
                    self['_rfm_perf_fns'][key] = value
                except KeyError:
                    raise ReframeSyntaxError(
                        f'the performance function {key!r} has already been '
                        f'defined in this class'
                    ) from None

            # Register the final methods
            if hasattr(value, '_rfm_final'):
                self['_rfm_final_methods'].add(key)

            # Register the hooks - if a value does not meet the conditions
            # it will be simply ignored
            self['_rfm_local_hook_registry'].add(value)

        def __getitem__(self, key):
            '''Expose and control access to the local namespaces.

            Variables may only be retrieved if their value has been previously
            set. Accessing a parameter in the class body is disallowed (the
            actual test parameter is set during the class instantiation).
            '''

            try:
                return super().__getitem__(key)
            except KeyError as err:
                try:
                    # Handle variable access
                    return self['_rfm_local_var_space'][key]
                except KeyError:
                    # Handle parameter access
                    if key in self['_rfm_local_param_space']:
                        raise ReframeSyntaxError(
                            'accessing a test parameter from the class '
                            'body is disallowed'
                        ) from None
                    elif key in self['_rfm_local_fixture_space']:
                        raise ReframeSyntaxError(
                            'accessing a fixture from the class body is '
                            'disallowed'
                        ) from None
                    else:
                        # As the last resource, look if key is a variable in
                        # any of the base classes. If so, make its value
                        # available in the current class' namespace.
                        for b in self['_rfm_bases']:
                            if key in b._rfm_var_space:
                                v = variables.ShadowVar(b._rfm_var_space[key])
                                self._namespace[key] = v
                                return v

                        # If 'key' is neither a variable nor a parameter,
                        # raise the exception from the base __getitem__.
                        raise err from None

        def reset(self, key):
            '''Reset an item to rerun it through the __setitem__ logic.'''
            self[key] = self[key]

    class WrappedFunction:
        '''Descriptor to wrap a free function as a bound-method.

        The free function object is wrapped by the constructor. Instances
        of this class should be inserted into the namespace of the target class
        with the desired name for the bound-method. Since this class is a
        descriptor, the `__get__` method will return the right bound-method
        when accessed from a class instance.

        :meta private:
        '''

        __slots__ = ('fn')

        def __init__(self, fn, name=None):
            @functools.wraps(fn)
            def _fn(*args, **kwargs):
                return fn(*args, **kwargs)

            self.fn = _fn
            if name:
                self.fn.__name__ = name

        def __get__(self, obj, objtype=None):
            if objtype is None:
                objtype = type(obj)

            self.fn.__qualname__ = '.'.join(
                [objtype.__qualname__, self.fn.__name__]
            )
            if obj is None:
                return self.fn

            return types.MethodType(self.fn, obj)

        def __call__(self, *args, **kwargs):
            return self.fn(*args, **kwargs)

        def __getattr__(self, name):
            if name in self.__slots__:
                return super().__getattr__(name)
            else:
                return getattr(self.fn, name)

        def __setattr__(self, name, value):
            if name in self.__slots__:
                super().__setattr__(name, value)
            else:
                setattr(self.fn, name, value)

    @classmethod
    def __prepare__(metacls, name, bases, **kwargs):
        namespace = super().__prepare__(name, bases, **kwargs)

        #
        # Initialize the various class level helper data structures
        #

        # Keep reference to the bases inside the namespace
        namespace['_rfm_bases'] = [
            b for b in bases if hasattr(b, '_rfm_var_space')
        ]

        # Regression test parameter space defined at the class level
        namespace['_rfm_local_param_space'] = namespaces.LocalNamespace()

        # Regression test var space defined at the class level
        namespace['_rfm_local_var_space'] = namespaces.LocalNamespace()

        # Regression test fixture space
        namespace['_rfm_local_fixture_space'] = namespaces.LocalNamespace()

        # Utility decorators
        namespace['_rfm_ext_bound'] = set()

        # Loggable properties
        namespace['_rfm_loggable_props'] = []

        namespace['_rfm_final_methods'] = set()
        namespace['_rfm_hook_registry'] = hooks.HookRegistry()
        namespace['_rfm_local_hook_registry'] = hooks.HookRegistry()
        namespace['_rfm_perf_fns'] = namespaces.LocalNamespace()

        def bind(fn, name=None):
            '''Directive to bind a free function to a class.

            See online docs for more information.

            .. note::
               Functions bound using this directive must be re-inspected after
               the class body execution has completed. This directive attaches
               the external method into the class namespace and returns the
               associated instance of the :class:`WrappedFunction`. However,
               this instance may be further modified by other ReFrame builtins
               such as :func:`run_before`, :func:`run_after`, :func:`final` and
               so on after it was added to the namespace, which would bypass
               the logic implemented in the :func:`__setitem__` method from the
               :class:`MetaNamespace` class. Hence, we track the items set by
               this directive in the ``_rfm_ext_bound`` set, so they can be
               later re-inspected.
            '''

            inst = metacls.WrappedFunction(fn, name)
            namespace[inst.__name__] = inst

            # Track the imported external functions
            namespace['_rfm_ext_bound'].add(inst.__name__)
            return inst

        # Register all builtins
        for name in builtins.__all__:
            namespace[name] = getattr(builtins, name)

        namespace['bind'] = bind
        return metacls.MetaNamespace(namespace)

    def __new__(metacls, name, bases, namespace, **kwargs):
        '''Remove builtins from the class namespace.

        It does not make sense to have the builtins available after the class
        was created or even at the instance level (e.g. doing
        ``self.parameter([1, 2, 3])`` does not make sense). So here, we
        intercept those builtins out of the namespace before the class is
        constructed.

        '''

        # Collect the loggable properties
        loggable_props = []
        namespace['_rfm_loggable_props'] = [
            v.fget._rfm_loggable for v in namespace.values()
            if hasattr(v, 'fget') and hasattr(v.fget, '_rfm_loggable')
        ]

        for n in builtins.__all__ + ['bind']:
            namespace.pop(n)

        # Reset the external functions imported through the bind directive.
        for item in namespace.pop('_rfm_ext_bound'):
            namespace.reset(item)

        return super().__new__(metacls, name, bases, dict(namespace), **kwargs)

    def __init__(cls, name, bases, namespace, **kwargs):
        super().__init__(name, bases, namespace, **kwargs)

        # Create a set with the attribute names already in use.
        cls._rfm_dir = set()
        for base in (b for b in bases if hasattr(b, '_rfm_dir')):
            cls._rfm_dir.update(base._rfm_dir)

        used_attribute_names = set(cls._rfm_dir).union(
            {h.__name__ for h in cls._rfm_local_hook_registry}
        )

        # Build the different global class namespaces
        namespace_types = (variables.VarSpace,
                           parameters.ParamSpace,
                           fixtures.FixtureSpace)
        for ns_type in namespace_types:
            ns = ns_type(cls, used_attribute_names)
            setattr(cls, ns.namespace_name, ns)
            used_attribute_names.update(ns.data())

        # Update used names set with the local __dict__
        cls._rfm_dir.update(cls.__dict__)

        # Populate the global hook registry with the hook registries of the
        # parent classes in reverse MRO order
        for c in list(reversed(cls.mro()))[:-1]:
            if hasattr(c, '_rfm_local_hook_registry'):
                cls._rfm_hook_registry.update(
                    c._rfm_local_hook_registry, denied_hooks=namespace
                )

        cls._rfm_hook_registry.update(cls._rfm_local_hook_registry)

        # Search the bases if no local sanity functions exist.
        if '_rfm_sanity' not in namespace:
            for base in cls._rfm_bases:
                if hasattr(base, '_rfm_sanity'):
                    cls._rfm_sanity = getattr(base, '_rfm_sanity')
                    if cls._rfm_sanity.__name__ in namespace:
                        raise ReframeSyntaxError(
                            f'{cls.__qualname__!r} overrides the candidate '
                            f'sanity function '
                            f'{cls._rfm_sanity.__qualname__!r} without '
                            f'defining an alternative'
                        )

                    break

        # Update the performance function dict with the bases.
        for base in cls._rfm_bases:
            for k, v in base._rfm_perf_fns.items():
                if k not in namespace:
                    try:
                        cls._rfm_perf_fns[k] = v
                    except KeyError:
                        '''Performance function overridden by other class'''

        # Add the final functions from its parents
        cls._rfm_final_methods.update(
            *(b._rfm_final_methods for b in cls._rfm_bases)
        )

        if getattr(cls, '_rfm_override_final', None):
            return

        for b in cls._rfm_bases:
            for key in b._rfm_final_methods:
                if key in namespace and callable(namespace[key]):
                    msg = (f"'{cls.__qualname__}.{key}' attempts to "
                           f"override final method "
                           f"'{b.__qualname__}.{key}'; "
                           f"you should use the pipeline hooks instead")
                    raise ReframeSyntaxError(msg)

    def __call__(cls, *args, **kwargs):
        '''Inject test builtins during object construction.

        When a class is instantiated, this method intercepts the arguments
        associated to the builtin  namespaces. This prevents both
        :func:`__new__` and :func:`__init__` methods from ever seing these
        arguments.

        The parameter and variable spaces are injected into the object after
        construction and before initialization.

        Fixtures must be injected after initialization. These are registered
        in the object (not the class) and they require certain attributes in
        the root test to be set before the fixture can be registered.

        :param variant_num: The test variant number. This must be an integer
          in the range of [0, cls.num_variants).
        :param variables: Mapping containing variables to be set during
          instantiation, but before initialization.
        '''

        # Intercept the requested variant number (if any) and map it to the
        # respective points in the parameter and fixture spaces.
        variant_num = kwargs.pop('variant_num', None)
        reset_sysenv = kwargs.pop('reset_sysenv', 0)
        param_index, fixt_index = cls._map_variant_num(variant_num)
        fixt_name = kwargs.pop('fixt_name', None)
        fixt_data = kwargs.pop('fixt_data', None)

        # Intercept variables to be set before initialization
        fixt_vars = kwargs.pop('fixt_vars', {})
        if not isinstance(fixt_vars, collections.abc.Mapping):
            raise TypeError("'fixt_vars' argument must be a mapping")

        obj = cls.__new__(cls, *args, **kwargs)

        # Insert the var and param spaces
        cls._rfm_var_space.inject(obj, cls)
        cls._rfm_param_space.inject(obj, cls, param_index)

        # Inject the variant numbers (if any present)
        if variant_num is not None:
            obj._rfm_variant_num = variant_num
            obj._rfm_param_variant = param_index
            obj._rfm_fixt_variant = fixt_index

        # Flag the instance as fixture
        if fixt_name:
            obj._rfm_unique_name = fixt_name
            obj._rfm_fixt_data = fixt_data
            obj._rfm_is_fixture = True
        else:
            obj._rfm_unique_name = cls.variant_name(variant_num)

        # Set the variables passed to the constructor
        for k, v in fixt_vars.items():
            if k in cls.var_space:
                setattr(obj, k, v)

        # Inject fixture proxies to give access to fixture parameters during
        # the init phase
        varinfo = cls.get_variant_info(variant_num, recurse=True)
        for fname, finfo in varinfo['fixtures'].items():
            fixt = cls.fixture_space[fname]
            if fixt.action != 'join':
                setattr(obj, fname, fixtures.FixtureProxy(finfo))

        obj.__init__(*args, **kwargs)
        if reset_sysenv & 1:
            obj.valid_systems = ['*']

        if reset_sysenv & 2:
            obj.valid_prog_environs = ['*']

        # Register the fixtures
        # Fixtures must be injected after the object's initialisation because
        # they require the valid_systems and valid_prog_environs attributes
        # from the object.
        cls._rfm_fixture_space.inject(obj, cls, fixt_index)
        return obj

    def __getattribute__(cls, name):
        '''Attribute lookup method for custom class attributes.

        ReFrame test variables are descriptors injected at the class level.
        If a variable descriptor has already been injected into the class,
        do not return the descriptor object and return the default value
        associated with that variable instead.

        .. warning::
            .. versionchanged:: 3.7.0
               Prior versions exposed the variable descriptor object if this
               was already present in the class, instead of returning the
               variable's default value.
        '''

        try:
            var_space = super().__getattribute__('_rfm_var_space')
        except AttributeError:
            var_space = None

        # If the variable is already injected, delegate lookup to __getattr__.
        if var_space and name in var_space.injected_vars:
            raise AttributeError('delegate variable lookup to __getattr__')

        # Default back to the base method if no special treatment required.
        return super().__getattribute__(name)

    def __getattr__(cls, name):
        '''Backup attribute lookup method into custom namespaces.

        Some ReFrame built-in types are stored under their own sub-namespaces.
        This method will perform an attribute lookup on these sub-namespaces
        if a call to the default :func:`__getattribute__` method fails to
        retrieve the requested class attribute.
        '''

        # Variable space lookup
        try:
            var_space = super().__getattribute__('_rfm_var_space')
            return var_space.vars[name]
        except AttributeError:
            '''Catch early access attempt to the variable space.'''
        except KeyError:
            '''Requested name not in variable space.'''

        # Parameter space lookup
        try:
            param_space = super().__getattribute__('_rfm_param_space')
            return param_space.params[name]
        except AttributeError:
            '''Catch early access attempt to the parameter space.'''
        except KeyError:
            '''Requested name not in parameter space.'''

        # Fixture space lookup
        try:
            fixture_space = super().__getattribute__('_rfm_fixture_space')
            return fixture_space.fixtures[name]
        except AttributeError:
            '''Catch early access attempt to the fixture space.'''
        except KeyError:
            '''Requested name not in fixture space.'''

        raise AttributeError(
            f'class {cls.__qualname__!r} has no attribute {name!r}'
        ) from None

    def setvar(cls, name, value):
        '''Set the value of a variable.

        :param name: The name of the variable.
        :param value: The value of the variable.

        :returns: :class:`True` if the variable was set.
            A variable will *not* be set, if it does not exist or when an
            attempt is made to set it with its underlying descriptor.
            This happens during the variable injection time and it should be
            delegated to the class' :func:`__setattr__` method.

        :raises ReframeSyntaxError: If an attempt is made to override a
            variable with a descriptor other than its underlying one.

        '''

        if '.' in name:
            # `name` refers to a fixture variable
            fixtname, varname = name.split('.', maxsplit=1)
            try:
                fixt_space = super().__getattribute__('_rfm_fixture_space')
            except AttributeError:
                '''Catch early access attempt to the variable space.'''

            if fixtname in fixt_space:
                return fixt_space[fixtname].cls.setvar(varname, value)

        try:
            var_space = super().__getattribute__('_rfm_var_space')
            if name in var_space:
                if not hasattr(value, '__get__'):
                    var_space[name].define(value)
                    return True
                elif var_space[name].field is not value:
                    desc = '.'.join([cls.__qualname__, name])
                    raise ReframeSyntaxError(
                        f'cannot override variable descriptor {desc!r}'
                    )
                else:
                    # Variable is being injected
                    return False
        except AttributeError:
            '''Catch early access attempt to the variable space.'''
            return False

    def __setattr__(cls, name, value):
        '''Handle the special treatment required for variables and parameters.

        A variable's default value can be updated when accessed as a regular
        class attribute. This behavior does not apply when the assigned value
        is a descriptor object. In that case, the task of setting the value is
        delegated to the base :func:`__setattr__` (this is to comply with
        standard Python behavior). However, since the variables are already
        descriptors which are injected during class instantiation, we disallow
        any attempt to override this descriptor (since it would be silently
        re-overridden in any case).

        Altering the value of a parameter when accessed as a class attribute
        is not allowed. This would break the parameter space internals.
        '''

        # Try to treat `name` as variable
        if cls.setvar(name, value):
            return

        # Try to treat `name` as a parameter
        try:
            # Catch attempts to override a test parameter
            param_space = super().__getattribute__('_rfm_param_space')
            if name in param_space.params:
                raise ReframeSyntaxError(f'cannot override parameter {name!r}')
        except AttributeError:
            '''Catch early access attempt to the parameter space.'''

        # Catch attempts to override a test fixture
        try:
            fixture_space = super().__getattribute__('_rfm_fixture_space')
            if name in fixture_space.fixtures:
                raise ReframeSyntaxError(f'cannot override fixture {name!r}')

        except AttributeError:
            '''Catch early access attempt to the fixture space.'''

        # Treat `name` as normal class attribute
        super().__setattr__(name, value)

    @property
    def num_variants(cls):
        '''Total number of variants of the test.'''
        return len(cls._rfm_param_space) * len(cls._rfm_fixture_space)

    def _map_variant_num(cls, variant_num=None):
        '''Map the global variant number into its sub-components.

        These are the coordinates for the parameter and fixture spaces.
        The parameter space index is the fast running one.
        '''

        if variant_num is None:
            return None, None

        # Bounds-check the variant number
        if variant_num >= cls.num_variants or variant_num < 0:
            raise ValueError(
                f'the provided variant number {variant_num} is out of bounds '
                f'[0, {cls.num_variants})'
            )

        p_space_size = len(cls._rfm_param_space)
        return variant_num % p_space_size, variant_num // p_space_size

    def get_variant_nums(cls, **conditions):
        '''Get the variant numbers that meet the specified conditions.

        The given conditions enable filtering the parameter space of the test.
        Filtering the fixture space is not allowed.

        .. code-block:: python

           # Filter out the test variants where my_param is greater than 3
           cls.get_variant_nums(my_param=lambda x: x < 4)

        The returned list of variant numbers can be passed to
        :func:`variant_name` in order to retrieve the actual test name.

        :param conditions: keyword arguments where the key is the test
            parameter name and the value is either a single value or a unary
            function that evaluates to :obj:`True` if the parameter point must
            be kept, :obj:`False` otherwise. If a single value is passed this
            is implicitly converted to the equality function, such that

            .. code-block:: python

               get_variant_nums(p=10)

            is equivalent to

            .. code-block:: python

               get_variant_nums(p=lambda x: x == 10)

        '''
        if not conditions:
            return list(range(cls.num_variants))

        # Filter the parameter indices - only for [0, len(cls.param_space))
        inner_variants = cls.param_space.get_variant_nums(**conditions)

        # The full variant space is of size param_space * fixture_space, and
        # the parameter space is the inner-most index in this coordinate
        # system. The inner_variants only contain the variants filtered in the
        # range of [0, len(cls.param_space)), so we use this "mask" to compute
        # the full variant indices in the range [0, cls.num_variants).
        outer_variants = []
        param_space_size = len(cls.param_space)
        for i in range(cls.num_variants // param_space_size):
            outer_variants += map(lambda x: x + param_space_size*i,
                                  inner_variants)

        return outer_variants

    def get_variant_info(cls, variant_num, *, recurse=False, max_depth=None,
                         **kwargs):
        '''Get the information from a given variant.

        This function returns a dictionary with the variant data on
        the different subspaces, such as the parameter values and the
        fixture variants.

        The parameter sub-dictionary contains the values for each parameter
        associated to the given variant number.

        The fixture sub-dictionary, by default, will return the variant number
        associated to each of the fixtures. However, if ``recurse`` is set to
        ``True``, each fixture entry will contain the full variant information
        for the given variant number. By default, the recursion will traverse
        the full fixture tree, but this recursion depth can be limited with the
        ``max_depth`` argument.

        See the example below:

        .. code:: python

          class Foo(rfm.RegressionTest):
              p0 = parameter(range(2))
              ...

          class Bar(rfm.RegressionTest):
              ...

          class MyTest(rfm.RegressionTest):
              p1 = parameter(['a', 'b'])
              f0 = fixture(Foo)
              f1 = fixture(Bar)
              ...

          # Get the raw info for variant 0
          MyTest.get_variant_info(0, recurse=True)
          # {
          #     'params': {'p1': 'a'},
          #     'fixtures': {
          #         'f0': {
          #             'params': {'p0': 0},
          #             'fixtures': {}
          #         },
          #         'f1': 0,
          #     }
          # }

        :param variant_num: An integer in the range of [0, cls.num_variants).
        :param recurse: Flag to control the recursion through the fixture
            space.
        :param max_depth: Set the recursion limit. When the ``recurse``
            argument is set to ``False``, this option has no effect.

        '''

        pid, fid = cls._map_variant_num(variant_num)
        ret = {
            'params': cls.param_space[pid] if pid is not None else {},
            'fixtures': cls.fixture_space[fid] if fid is not None else {}
        }

        # Get current recursion level
        rdepth = kwargs.get('_current_depth', 0)
        if recurse and (max_depth is None or rdepth < max_depth):
            for fname, variant in ret['fixtures'].items():
                if len(variant) > 1:
                    continue

                fixt = cls.fixture_space[fname]
                ret['fixtures'][fname] = fixt.cls.get_variant_info(
                    variant[0], recurse=recurse, max_depth=max_depth,
                    _current_depth=rdepth+1
                )

        return ret

    @property
    def raw_params(cls):
        '''Expose the raw parameters.'''
        return cls.param_space.params

    @property
    def param_space(cls):
        '''Expose the parameter space.'''
        return cls._rfm_param_space

    @property
    def var_space(cls):
        '''Expose the variable space.'''
        return cls._rfm_var_space

    @property
    def fixture_space(cls):
        '''Expose the fixture space.'''
        return cls._rfm_fixture_space

    def is_abstract(cls):
        '''Check if the class is an abstract test.

        A test is considered abstract if any of its direct or indirect
        parameters (inherited from a base class or from a fixture) is
        undefined.

        :returns: :obj:`True` if the test is abstract, :obj:`False` otherwise.

        '''
        return cls.num_variants == 0

    def variant_name(cls, variant_num=None):
        '''Return the name of the test variant with a specific variant number.

        :param variant_num: An integer in the range of ``[0, cls.num_variants)``.
        '''

        name = cls.__name__
        if variant_num is None:
            return name

        if cls.num_variants > 1:
            width = utils.count_digits(cls.num_variants)
            name += f'_{variant_num:0{width}}'

        return name

    def loggable_attrs(cls):
        '''Get the loggable attributes of this class.'''
        loggable_vars = [(name, None) for name, var in cls.var_space.items()
                         if var.is_loggable()]
        loggable_params = [
            (name, None) for name, param in cls.param_space.items()
            if param.is_loggable()
        ]
        loggable_props = []
        for c in cls.mro():
            if hasattr(c, '_rfm_loggable_props'):
                loggable_props += c._rfm_loggable_props

        return sorted(loggable_props + loggable_vars + loggable_params)


def make_test(name, bases, body, methods=None, **kwargs):
    '''Define a new test class programmatically.

    Using this method is completely equivalent to using the :keyword:`class`
    to define the test class. More specifically, the following:

    .. code-block:: python

       hello_cls = rfm.make_test(
           'HelloTest', (rfm.RunOnlyRegressionTest,),
           {
               'valid_systems': ['*'],
               'valid_prog_environs': ['*'],
               'executable': 'echo',
               'sanity_patterns': sn.assert_true(1)
           }
       )

    is completely equivalent to

    .. code-block:: python

       class HelloTest(rfm.RunOnlyRegressionTest):
           valid_systems = ['*']
           valid_prog_environs = ['*']
           executable = 'echo'
           sanity_patterns: sn.assert_true(1)

       hello_cls = HelloTest

    Test :ref:`builtins <builtins>` can also be used when defining the body of
    the test by accessing them through the :obj:`reframe.core.builtins`.
    Methods can also be bound to the newly created tests using the ``methods``
    argument. The following is an example:

    .. code-block:: python

       import reframe.core.builtins as builtins


       def set_message(obj):
           obj.executable_opts = [obj.message]

       def validate(obj):
           return sn.assert_found(obj.message, obj.stdout)

       hello_cls = rfm.make_test(
           'HelloTest', (rfm.RunOnlyRegressionTest,),
           {
               'valid_systems': ['*'],
               'valid_prog_environs': ['*'],
               'executable': 'echo',
               'message': builtins.variable(str)
           },
           methods=[
               builtins.run_before('run')(set_message),
               builtins.sanity_function(validate)
           ]
       )


    :param name: The name of the new test class.
    :param bases: A tuple of the base classes of the class that is being
        created.
    :param body: A mapping of key/value pairs that will be inserted as class
        attributes in the newly created class.
    :param methods: A list of functions to be bound as methods to the class
        that is being created. The functions will be bound with their original
        name.
    :param kwargs: Any keyword arguments to be passed to the
        :class:`RegressionTestMeta` metaclass.

    .. versionadded:: 3.10.0

    .. versionchanged:: 3.11.0
       Added the ``methods`` arguments.

    '''
    namespace = RegressionTestMeta.__prepare__(name, bases, **kwargs)
    methods = methods or []

    # Add methods to the body
    for m in methods:
        body[m.__name__] = m

    # We update the namespace with the body of the class and we explicitly
    # call reset on each namespace key to trigger the functionality of
    # `__setitem__()` as if the body elements were actually being typed in the
    # class definition
    namespace.update(body)
    for k in list(namespace.keys()):
        namespace.reset(k)

    cls = RegressionTestMeta(name, bases, namespace, **kwargs)
    return cls



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import json
import jsonschema
import os
import shutil
import tempfile
import traceback

import reframe as rfm
import reframe.core.runtime as runtime
import reframe.utility.osext as osext
from reframe.core.exceptions import ConfigError
from reframe.core.logging import getlogger
from reframe.core.schedulers import Job
from reframe.core.systems import DeviceInfo, ProcessorInfo
from reframe.utility.cpuinfo import cpuinfo


# This is meant to be used by the unit tests
_TREAT_WARNINGS_AS_ERRORS = False


def _contents(filename):
    '''Return the contents of a file.'''

    with open(filename) as fp:
        return fp.read()


def _log_contents(filename):
    filename = os.path.abspath(filename)
    getlogger().debug(f'--- {filename} ---\n'
                      f'{_contents(filename)}\n'
                      f'--- {filename} ---')


class _copy_reframe:
    def __init__(self, prefix):
        self._prefix = prefix
        self._workdir = None

    def __enter__(self):
        self._workdir = os.path.abspath(
            tempfile.mkdtemp(prefix='rfm.', dir=self._prefix)
        )
        paths = ['bin/', 'reframe/', 'bootstrap.sh', 'requirements.txt']
        for p in paths:
            src = os.path.join(rfm.INSTALL_PREFIX, p)
            if os.path.isdir(src):
                dst = os.path.join(self._workdir, p)
                osext.copytree(src, dst, dirs_exist_ok=True)
            else:
                shutil.copy2(src, self._workdir)

        return self._workdir

    def __exit__(self, exc_type, exc_val, exc_tb):
        osext.rmtree(self._workdir)


def _subschema(fragment):
    '''Create a configuration subschema.'''

    full_schema = runtime.runtime().site_config.schema
    return {
        '$schema': full_schema['$schema'],
        'defs': full_schema['defs'],
        '$ref': fragment
    }


def _validate_info(info, schema):
    if schema is None:
        return info

    jsonschema.validate(info, schema)
    return info


def _load_info(filename, schema=None):
    try:
        with open(filename) as fp:
            return _validate_info(json.load(fp), schema)
    except OSError as e:
        if _TREAT_WARNINGS_AS_ERRORS:
            raise

        getlogger().warning(
            f'could not load file: {filename!r}: {e}'
        )
        return {}
    except jsonschema.ValidationError as e:
        raise ConfigError(
            f'could not validate meta-config file {filename!r}'
        ) from e


def _save_info(filename, topo_info):
    if not topo_info:
        return

    os.makedirs(os.path.dirname(filename), exist_ok=True)
    try:
        with open(filename, 'w') as fp:
            json.dump(topo_info, fp, indent=2)
    except OSError as e:
        if _TREAT_WARNINGS_AS_ERRORS:
            raise

        getlogger().warning(
            f'could not save topology file: {filename!r}: {e}'
        )
    else:
        getlogger().debug(f'> saved topology in {filename!r}')


def _is_part_local(part):
    return (part.scheduler.registered_name == 'local' and
            part.launcher_type.registered_name == 'local')


def _remote_detect(part):
    def _emit_script(job, env):
        launcher_cmd = job.launcher.run_command(job)
        commands = [
            f'./bootstrap.sh',
            f'{launcher_cmd} ./bin/reframe --detect-host-topology=topo.json'
        ]
        job.prepare(commands, env, trap_errors=True)

    getlogger().info(
        f'Detecting topology of remote partition {part.fullname!r}: '
        f'this may take some time...'
    )
    topo_info = {}
    try:
        prefix = runtime.runtime().get_option('general/0/remote_workdir')
        with _copy_reframe(prefix) as dirname:
            with osext.change_dir(dirname):
                job = Job.create(part.scheduler,
                                 part.launcher_type(),
                                 name='rfm-detect-job',
                                 sched_access=part.access)
                _emit_script(job, [part.local_env])
                getlogger().debug('submitting detection script')
                _log_contents(job.script_filename)
                job.submit()
                job.wait()
                getlogger().debug('job finished')
                _log_contents(job.stdout)
                _log_contents(job.stderr)
                topo_info = json.loads(_contents('topo.json'))
    except Exception as e:
        if _TREAT_WARNINGS_AS_ERRORS:
            raise

        getlogger().warning(f'failed to retrieve remote processor info: {e}')
        getlogger().debug(traceback.format_exc())

    return topo_info


def detect_topology():
    rt = runtime.runtime()
    detect_remote_systems = rt.get_option('general/0/remote_detect')
    topo_prefix = os.path.join(os.getenv('HOME'), '.reframe/topology')
    for part in rt.system.partitions:
        getlogger().debug(f'detecting topology info for {part.fullname}')
        found_procinfo = False
        found_devinfo  = False
        if part.processor.info != {}:
            # Processor info set up already in the configuration
            getlogger().debug(
                f'> topology found in configuration file; skipping...'
            )
            found_procinfo = True

        if part.devices:
            # Devices set up already in the configuration
            getlogger().debug(
                f'> devices found in configuration file; skipping...'
            )
            found_devinfo = True

        if found_procinfo and found_devinfo:
            continue

        topo_file = os.path.join(
            topo_prefix, f'{rt.system.name}-{part.name}', 'processor.json'
        )
        dev_file = os.path.join(
            topo_prefix, f'{rt.system.name}-{part.name}', 'devices.json'
        )
        if not found_procinfo and os.path.exists(topo_file):
            getlogger().debug(
                f'> found topology file {topo_file!r}; loading...'
            )
            try:
                part._processor = ProcessorInfo(
                    _load_info(topo_file, _subschema('#/defs/processor_info'))
                )
                found_procinfo = True
            except json.decoder.JSONDecodeError as e:
                getlogger().debug(
                    f'> could not load {topo_file!r}: {e}: ignoring...'
                )

        if not found_devinfo and os.path.exists(dev_file):
            getlogger().debug(
                f'> found devices file {dev_file!r}; loading...'
            )
            try:
                devices_info = _load_info(
                    dev_file, _subschema('#/defs/devices')
                )
                part._devices = [DeviceInfo(d) for d in devices_info]
                found_devinfo = True
            except json.decoder.JSONDecodeError as e:
                getlogger().debug(
                    f'> could not load {dev_file!r}: {e}: ignoring...'
                )

        if found_procinfo and found_devinfo:
            continue

        if not found_procinfo:
            # No topology found, try to auto-detect it
            getlogger().debug(f'> no topology file found; auto-detecting...')
            modules = list(rt.system.preload_environ.modules)
            vars = dict(rt.system.preload_environ.env_vars.items())
            if _is_part_local(part):
                modules += part.local_env.modules
                vars.update(part.local_env.env_vars)

                # Unconditionally detect the system for fully local partitions
                with runtime.temp_environment(modules=modules, env_vars=vars):
                    part._processor = ProcessorInfo(cpuinfo())

                _save_info(topo_file, part.processor.info)
            elif detect_remote_systems:
                with runtime.temp_environment(modules=modules, env_vars=vars):
                    part._processor = ProcessorInfo(_remote_detect(part))

                if part.processor.info:
                    _save_info(topo_file, part.processor.info)

        if not found_devinfo:
            getlogger().debug(f'> device auto-detection is not supported')



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import functools
import inspect

import reframe.utility as util


def attach_to(phase):
    '''Backend function to attach a hook to a given phase.

    :meta private:
    '''
    def deco(func):
        if hasattr(func, '_rfm_attach'):
            func._rfm_attach.append(phase)
        else:
            func._rfm_attach = [phase]

        try:
            # no need to resolve dependencies independently; this function is
            # already attached to a different phase
            func._rfm_resolve_deps = False
        except AttributeError:
            pass

        @functools.wraps(func)
        def _fn(*args, **kwargs):
            func(*args, **kwargs)

        return _fn

    return deco


def require_deps(func):
    '''Decorator to denote that a function will use the test dependencies.

    The arguments of the decorated function must be named after the
    dependencies that the function intends to use. The decorator will bind the
    arguments to a partial realization of the
    :func:`~reframe.core.pipeline.RegressionTest.getdep` function, such that
    conceptually the new function arguments will be the following:

    .. code-block:: python

       new_arg = functools.partial(getdep, orig_arg_name)

    The converted arguments are essentially functions accepting a single
    argument, which is the target test's programming environment.
    Additionally, this decorator will attach the function to run *after* the
    test's setup phase, but *before* any other "post-setup" pipeline hook.

    .. warning::
       .. versionchanged:: 3.7.0
          Using this functionality from the :py:mod:`reframe` or
          :py:mod:`reframe.core.decorators` modules is now deprecated. You
          should use the built-in function described here.

       .. versionchanged:: 4.0.0
          You may only use this function as framework built-in.

    '''

    tests = inspect.getfullargspec(func).args[1:]
    func._rfm_resolve_deps = True

    @functools.wraps(func)
    def _fn(obj, *args):
        newargs = [functools.partial(obj.getdep, t) for t in tests]
        func(obj, *newargs)

    return _fn


def attach_hooks(hooks):
    '''Attach pipeline hooks to phase ``name''.

    This function returns a decorator for pipeline functions that will run the
    registered hooks before and after the function.

    If ``name'' is :class:`None`, both pre- and post-hooks will run, otherwise
    only the hooks of the phase ``name'' will be executed.
    '''

    def _deco(func):
        def select_hooks(obj, kind):
            phase = kind + func.__name__
            if phase not in hooks:
                return []

            return [h for h in hooks.get(phase, [])
                    if h.__name__ not in getattr(obj, '_disabled_hooks', [])]

        @functools.wraps(func)
        def _fn(obj, *args, **kwargs):
            for h in select_hooks(obj, 'pre_'):
                getattr(obj, h.__name__)()

            func(obj, *args, **kwargs)
            for h in select_hooks(obj, 'post_'):
                getattr(obj, h.__name__)()

        return _fn

    return _deco


class Hook:
    '''A pipeline hook.

    This is essentially a function wrapper that hashes the functions by name,
    since we want hooks to be overriden by name in subclasses.
    '''

    def __init__(self, fn):
        self.__fn = fn
        if not hasattr(fn, '_rfm_attach'):
            raise ValueError(f'{fn.__name__} is not a hook')

    @property
    def stages(self):
        return self._rfm_attach

    def __getattr__(self, attr):
        return getattr(self.__fn, attr)

    @property
    def fn(self):
        return self.__fn

    def __hash__(self):
        return hash(self.__name__)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented

        return self.__name__ == other.__name__

    def __call__(self, *args, **kwargs):
        return self.__fn(*args, **kwargs)

    def __repr__(self):
        return repr(self.__fn)


class HookRegistry:
    '''Global hook registry.'''

    def __init__(self, hooks=None):
        self.__hooks = util.OrderedSet()
        if hooks is not None:
            self.update(hooks)

    def __contains__(self, key):
        return key in self.__hooks

    def __getattr__(self, name):
        return getattr(self.__hooks, name)

    def __iter__(self):
        return iter(self.__hooks)

    def add(self, v):
        '''Add value to the hook registry if it meets the conditions.

        Hook functions have an `_rfm_attach` attribute that specify the stages
        of the pipeline where they must be attached. Dependencies will be
        resolved first in the post-setup phase if not assigned elsewhere.
        '''

        if hasattr(v, '_rfm_attach'):
            # Always override hooks with the same name
            h = Hook(v)
            self.__hooks.discard(h)
            self.__hooks.add(h)
        elif hasattr(v, '_rfm_resolve_deps'):
            v._rfm_attach = ['post_setup']
            self.__hooks.add(Hook(v))

    def update(self, other, *, denied_hooks=None):
        '''Update the hook registry with the hooks from another hook registry.

        The optional ``denied_hooks`` argument takes a set of disallowed
        hook names, preventing their inclusion into the current hook registry.
        '''

        assert isinstance(other, HookRegistry)
        denied_hooks = denied_hooks or set()
        for h in other:
            if h.__name__ not in denied_hooks:
                # Hooks in `other` override ours
                self.__hooks.discard(h)
                self.__hooks.add(h)

    def __repr__(self):
        return repr(self.__hooks)



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#
# Decorators used for the definition of tests
#

__all__ = ['simple_test']


import inspect
import sys
import traceback

import reframe.utility.osext as osext
from reframe.core.exceptions import ReframeSyntaxError, SkipTestError, what
from reframe.core.fixtures import FixtureRegistry
from reframe.core.logging import getlogger, time_function
from reframe.core.pipeline import RegressionTest
from reframe.utility.versioning import VersionValidator


# NOTE: we should consider renaming this module in 4.0; it practically takes
# care of the registration and instantiation of the tests.


class TestRegistry:
    '''Regression test registry.

    The tests are stored in a dictionary where the test class is the key
    and the constructor arguments for the different instantiations of the
    test are stored as the dictionary value as a list of (args, kwargs)
    tuples.

    For backward compatibility reasons, the registry also contains a set of
    tests to be skipped. The machinery related to this should be dropped with
    the ``required_version`` decorator.
    '''

    def __init__(self):
        self._tests = dict()

    @classmethod
    def create(cls, test, *args, **kwargs):
        obj = cls()
        obj.add(test, *args, **kwargs)
        return obj

    def add(self, test, *args, **kwargs):
        self._tests.setdefault(test, [])
        self._tests[test].append((args, kwargs))

    @time_function
    def instantiate_all(self, reset_sysenv=0):
        '''Instantiate all the registered tests.

        :param reset_sysenv: Reset valid_systems and valid_prog_environs after
            instantiating the tests. Bit 0 resets the valid_systems, bit 1
            resets the valid_prog_environs.
        '''

        # We first instantiate the leaf tests and then walk up their
        # dependencies to instantiate all the fixtures. Fixtures can only
        # establish their exact dependencies at instantiation time, so the
        # dependency graph grows dynamically.

        leaf_tests = []
        for test, variants in self._tests.items():
            for args, kwargs in variants:
                try:
                    kwargs['reset_sysenv'] = reset_sysenv
                    leaf_tests.append(test(*args, **kwargs))
                except SkipTestError as e:
                    getlogger().warning(
                        f'skipping test {test.__qualname__!r}: {e}'
                    )
                except Exception:
                    exc_info = sys.exc_info()
                    getlogger().warning(
                        f"skipping test {test.__qualname__!r}: "
                        f"{what(*exc_info)} "
                        f"(rerun with '-v' for more information)"
                    )
                    getlogger().verbose(traceback.format_exc())

        # Instantiate fixtures

        # Do a level-order traversal of the fixture registries of all leaf
        # tests, instantiate all fixtures and generate the final set of
        # candidate tests; the leaf tests are consumed at the end of the
        # traversal and all instantiated tests (including fixtures) are stored
        # in `final_tests`.
        final_tests = []
        fixture_registry = FixtureRegistry()
        while leaf_tests:
            tmp_registry = FixtureRegistry()
            while leaf_tests:
                c = leaf_tests.pop()
                reg = getattr(c, '_rfm_fixture_registry', None)
                final_tests.append(c)
                if reg:
                    tmp_registry.update(reg)

            # Instantiate the new fixtures and update the registry
            new_fixtures = tmp_registry.difference(fixture_registry)
            leaf_tests = new_fixtures.instantiate_all()
            fixture_registry.update(new_fixtures)

        return final_tests

    def __iter__(self):
        '''Iterate over the registered test classes.'''
        return iter(self._tests.keys())

    def __contains__(self, test):
        return test in self._tests


def _register_test(cls, *args, **kwargs):
    '''Register a test and its construction arguments into the registry.'''

    mod = inspect.getmodule(cls)
    if not hasattr(mod, '_rfm_test_registry'):
        mod._rfm_test_registry = TestRegistry.create(cls, *args, **kwargs)
    else:
        mod._rfm_test_registry.add(cls, *args, **kwargs)


def _validate_test(cls):
    if not issubclass(cls, RegressionTest):
        raise ReframeSyntaxError('the decorated class must be a '
                                 'subclass of RegressionTest')

    if (cls.is_abstract()):
        getlogger().warning(
            f'skipping test {cls.__qualname__!r}: '
            f'test has one or more undefined parameters'
        )
        return False

    conditions = [VersionValidator(v) for v in cls._rfm_required_version]
    if (cls._rfm_required_version and
        not any(c.validate(osext.reframe_version()) for c in conditions)):

        getlogger().warning(f"skipping incompatible test "
                            f"'{cls.__qualname__}': not valid for ReFrame "
                            f"version {osext.reframe_version().split('-')[0]}")
        return False

    return True


def simple_test(cls):
    '''Class decorator for registering tests with ReFrame.

    The decorated class must derive from
    :class:`reframe.core.pipeline.RegressionTest`. This decorator is also
    available directly under the :mod:`reframe` module.

    .. versionadded:: 2.13
    '''
    if _validate_test(cls):
        for n in range(cls.num_variants):
            _register_test(cls, variant_num=n)

    return cls



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#
# Flux-Framework backend
#
# - Initial version submitted by Vanessa Sochat,
#   Lawrence Livermore National Lab
#

import itertools
import os
import time

from reframe.core.backends import register_scheduler
from reframe.core.exceptions import JobError
from reframe.core.schedulers import JobScheduler, Job

# Just import flux once
try:
    import flux
    import flux.job
    from flux.job import JobspecV1
except ImportError:
    error = 'no flux Python bindings found'
else:
    error = None

WAITING_STATES = ('QUEUED', 'HELD', 'WAITING', 'PENDING')


class _FluxJob(Job):
    def __init__(self, *args, **kwargs):
        '''Create the flux job (and future) to watch.'''
        super().__init__(*args, **kwargs)

        # Generate the flux job
        self.fluxjob = JobspecV1.from_command(
            command=['/bin/bash', self.script_filename],
            num_tasks=self.num_tasks_per_core or 1,
            cores_per_task=self.num_cpus_per_task or 1,
        )

        # We must use absolute paths for Flux
        out = os.path.join(os.path.abspath(self.workdir), self.stdout)
        err = os.path.join(os.path.abspath(self.workdir), self.stderr)

        # A duration of zero (the default) means unlimited
        self.fluxjob.duration = self.time_limit or 0
        self.fluxjob.stdout = out
        self.fluxjob.stderr = err
        self.fluxjob.cwd = os.path.abspath(self.workdir)
        self.fluxjob.environment = dict(os.environ)
        self._completed = False

    @property
    def completed(self):
        return self._completed


@register_scheduler('flux', error=error)
class FluxJobScheduler(JobScheduler):
    def __init__(self):
        self._fexecutor = flux.job.FluxExecutor()
        self._submit_timeout = self.get_option('job_submit_timeout')

    def emit_preamble(self, job):
        # We don't need to submit with a file, so we don't need a preamble.
        return []

    def make_job(self, *args, **kwargs):
        return _FluxJob(*args, **kwargs)

    def submit(self, job):
        '''Submit a job to the flux executor.'''

        flux_future = self._fexecutor.submit(job.fluxjob)
        job._jobid = str(flux_future.jobid())
        job._submit_time = time.time()
        job._flux_future = flux_future

    def cancel(self, job):
        '''Cancel a running Flux job.'''

        # Job future cannot cancel once running or completed
        if not job._flux_future.cancel():
            # This will raise JobException with event=cancel (on poll)
            flux.job.cancel(flux.Flux(), job._flux_future.jobid())

    def poll(self, *jobs):
        '''Poll running Flux jobs for updated states.'''

        if jobs:
            # filter out non-jobs
            jobs = [job for job in jobs if job is not None]

        if not jobs:
            return

        # Loop through active jobs and act on status
        for job in jobs:
            if job._flux_future.done():
                try:
                    # The exit code can help us determine if the job was
                    # successful
                    exit_code = job._flux_future.result(0)
                except flux.job.JobException:
                    # Currently the only state we see is cancelled here
                    self.log(f'Job {job.jobid} was likely cancelled.')
                    job._state = 'CANCELLED'
                except RuntimeError:
                    # Assume some runtime issue (suspended)
                    self.log(f'Job {job.jobid} was likely suspended.')
                    job._state = 'SUSPENDED'
                else:
                    # the job finished (but possibly with nonzero exit code)
                    job._state = 'COMPLETED'
                    if exit_code != 0:
                        self.log(
                            f'Job {job.jobid} did not finish successfully'
                        )

                job._completed = True
            elif job.state in WAITING_STATES and job.max_pending_time:
                if time.time() - job.submit_time >= job.max_pending_time:
                    self.cancel(job)
                    job._exception = JobError(
                        'maximum pending time exceeded', job.jobid
                    )
            else:
                # Otherwise, we are still running
                job._state = 'RUNNING'

    def allnodes(self):
        raise NotImplementedError('flux backend does not support node listing')

    def filternodes(self, job, nodes):
        raise NotImplementedError(
            'flux backend does not support node filtering'
        )

    def wait(self, job):
        '''Wait until a job is finished.'''

        intervals = itertools.cycle([1, 2, 3])
        while not self.finished(job):
            self.poll(job)
            time.sleep(next(intervals))

    def finished(self, job):
        if job.exception:
            raise job.exception

        return job.completed



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#
# Functionality to build extensible parameterized tests.
#

import copy
import itertools

import reframe.core.namespaces as namespaces
import reframe.utility as utils
from reframe.core.exceptions import ReframeSyntaxError


class TestParam:
    '''Inserts a new test parameter.

    At the class level, these parameters are stored in a separate namespace
    referred to as the *parameter space*. If a parameter with a matching name
    is already present in the parameter space of a parent class, the existing
    parameter values will be combined with those provided by this method
    following the inheritance behavior set by the arguments ``inherit_params``
    and ``filter_params``. Instead, if no parameter with a matching name
    exists in any of the parent parameter spaces, a new regression test
    parameter is created. A regression test can be parameterized as follows:

    .. code:: python

        class Foo(rfm.RegressionTest):
            variant = parameter(['A', 'B'])

            # print(variant)
            # Error: a parameter may only be accessed from the class instance

            @run_after('init')
            def do_something(self):
                if self.variant == 'A':
                    do_this()
                else:
                    do_other()

    One of the most powerful features of these built-in functions is that they
    store their input information at the class level. However, a parameter may
    only be accessed from the class instance and accessing it directly from
    the class body is disallowed. With this approach, extending or
    specializing an existing parameterized regression test becomes
    straightforward, since the test attribute additions and modifications made
    through built-in functions in the parent class are automatically inherited
    by the child test. For instance, continuing with the example above, one
    could override the :func:`do_something` hook in the :class:`Foo`
    regression test as follows:

    .. code:: python

       class Bar(Foo):
           @run_after('init')
           def do_something(self):
               if self.variant == 'A':
                   override_this()
               else:
                   override_other()

    Moreover, a derived class may extend, partially extend and/or modify the
    parameter values provided in the base class as shown below.

    .. code:: python

       class ExtendVariant(Bar):
           # Extend the full set of inherited variant parameter values
           # to ['A', 'B', 'C']
           variant = parameter(['C'], inherit_params=True)

       class PartiallyExtendVariant(Bar):
           # Extend a subset of the inherited variant parameter values
           # to ['A', 'D']
           variant = parameter(['D'], inherit_params=True,
                               filter_params=lambda x: x[:1])

       class ModifyVariant(Bar):
           # Modify the variant parameter values to ['AA', 'BA']
           variant = parameter(inherit_params=True,
                              filter_params=lambda x: map(lambda y: y+'A', x))

    A parameter with no values is referred to as an *abstract parameter* (i.e.
    a parameter that is declared but not defined). Therefore, classes with at
    least one abstract parameter are considered abstract classes.

    .. code:: python

       class AbstractA(Bar):
           variant = parameter()

       class AbstractB(Bar):
           variant = parameter(inherit_params=True, filter_params=lambda x: [])

    :param values: An iterable containing the parameter values.

    :param inherit_params: If :obj:`True`, the parameter values defined in any
        base class will be inherited. In this case, the parameter values
        provided in the current class will extend the set of inherited
        parameter values. If the parameter does not exist in any of the parent
        parameter spaces, this option has no effect.

    :param filter_params: Function to filter/modify the inherited parameter
        values that may have been provided in any of the parent parameter
        spaces. This function must accept a single iterable argument and
        return an iterable. It will be called with the inherited parameter
        values and it must return the filtered set of parameter values. This
        function will only have an effect if used with
        ``inherit_params=True``.

    :param fmt: A formatting function that will be used to format the values
        of this parameter in the test's
        :attr:`~reframe.core.pipeline.RegressionTest.display_name`. This
        function should take as argument the parameter value and return a
        string representation of the value. If the returned value is not a
        string, it will be converted using the :py:func:`str` function.

    :param loggable: Mark this parameter as loggable. If :obj:`True`, this
        parameter will become a log record attribute under the name
        ``check_NAME``, where ``NAME`` is the name of the parameter.

    :returns: A new test parameter.

    .. versionadded:: 3.10.0
       The ``fmt`` argument is added.

    .. versionadded:: 3.11.0
       The ``loggable`` argument is added.
    '''

    def __init__(self, values=None, inherit_params=False,
                 filter_params=None, fmt=None, loggable=False):
        if values is None:
            values = []

        if not inherit_params:
            # By default, filter out all the parameter values defined in the
            # base classes.
            def filter_params(x):
                return ()
        elif filter_params is None:
            # If inherit_params==True, inherit all the parameter values from
            # the base classes as default behaviour.
            def filter_params(x):
                return x

        self.values = tuple(values)

        # Validate and set the filter_params function
        if (not callable(filter_params) or
            not utils.is_trivially_callable(filter_params, non_def_args=1)):
            raise TypeError("'filter_params' argument must be a callable "
                            "accepting a single argument")

        self.filter_params = filter_params

        # Validate and set the alternative function
        if fmt is None:
            def fmt(x):
                return x

        if (not callable(fmt) or
            not utils.is_trivially_callable(fmt, non_def_args=1)):
            raise TypeError("'fmt' argument must be a callable "
                            "accepting a single argument")

        self.__fmt_fn = fmt
        self.__loggable = loggable

    @property
    def format(self):
        return self.__fmt_fn

    def update(self, other):
        '''Update this parameter from another one.

        The values from the other parameter will be filtered according to the
        filter function of this one and prepended to this parameter's values.

        :meta private:
        '''

        try:
            filt_vals = self.filter_params(other.values)
        except Exception:
            raise
        else:
            try:
                self.values = tuple(filt_vals) + self.values
            except TypeError:
                raise ReframeSyntaxError(
                    f"'filter_param' must return an iterable"
                ) from None

    def is_abstract(self):
        return len(self.values) == 0

    def is_loggable(self):
        return self.__loggable


class ParamSpace(namespaces.Namespace):
    '''Regression test parameter space

    The parameter space is stored as a dictionary (self.params), where the
    keys are the parameter names and the values are tuples with all the
    available values for each parameter. The __init__ method in this class
    takes the optional argument ``target_cls``, which is the regression test
    class that the parameter space is being built for. If no target class is
    provided, the parameter space is initialized as empty.

    All the parameter combinations are stored under ``__param_combinations``.
    This enables random-access to any of the available parameter combinations
    through the ``__getitem__`` method.
    '''

    def __init__(self, target_cls=None, illegal_names=None):
        super().__init__(target_cls, illegal_names,
                         ns_name='_rfm_param_space',
                         ns_local_name='_rfm_local_param_space')

        # Store all param combinations to allow random access.
        self.__param_combinations = tuple(
            itertools.product(
                *(copy.deepcopy(p.values) for p in self.params.values())
            )
        )

        # Map the parameter names to the position they are stored in the
        # parameter space
        self._position = {name: idx for idx, name in enumerate(self.params)}

    def join(self, other, cls):
        '''Join other parameter space into the current one.

        Join two different parameter spaces into a single one. Both parameter
        spaces must be an instance ot the ParamSpace class. This method will
        raise an error if a parameter is defined in the two parameter spaces
        to be merged.

        :param other: instance of the ParamSpace class.
        :param cls: the target class.
        '''
        for name in other.params:
            # With multiple inheritance, a single parameter
            # could be doubly defined and lead to repeated
            # values
            if self.defines(name) and other.defines(name):
                raise ReframeSyntaxError(
                    f'parameter space conflict: '
                    f'parameter {name!r} is defined in more than '
                    f'one base class of class {cls.__qualname__!r}'
                )

            if not self.defines(name):
                # If we do not define the parameter, take it from other
                self.params[name] = other.params[name]

    def extend(self, cls):
        '''Extend the parameter space with the local parameter space.'''

        local_param_space = getattr(cls, self.local_namespace_name, dict())
        for name, p in local_param_space.items():
            if name in self.params:
                p.update(self.params[name])

            self.params[name] = p

        # Clear the local param space
        local_param_space.clear()

        # If any previously declared parameter was defined in the class body
        # by directly assigning it a value, raise an error. Parameters must be
        # changed using the `x = parameter([...])` syntax.
        for key, values in cls.__dict__.items():
            if key in self.params:
                raise ReframeSyntaxError(
                    f'parameter {key!r} must be modified through the built-in '
                    f'parameter type'
                )

    def inject(self, obj, cls=None, params_index=None):
        '''Insert the params in the regression test.

        Create and initialize the regression test parameters as object
        attributes. The values assigned to these parameters exclusively depend
        on the value of params_index. This argument is simply an index to a
        a given parametere combination. If params_index is left with its
        default value, the regression test parameters are initialized as
        None.

        :param obj: The test object.
        :param cls: The test class.
        :param params_index: index to a point in the parameter space.
        '''
        # Set the values of the test parameters (if any)
        if self.params and params_index is not None:
            try:
                # Get the parameter values for the specified variant
                param_values = self.__param_combinations[params_index]
            except IndexError as no_params:
                raise RuntimeError(
                    f'parameter space index out of range for '
                    f'{obj.__class__.__qualname__}'
                ) from None
            else:
                for index, key in enumerate(self.params):
                    setattr(obj, key, param_values[index])

        else:
            # Otherwise init the params as None
            for key in self.params:
                setattr(obj, key, None)

    @property
    def params(self):
        return self._namespace

    def defines(self, name):
        '''Return True if parameter is defined.

        A parameter is defined if it exists in the namespace and it is not
        abstract.
        '''
        return name in self.params and not self.params[name].is_abstract()

    def __iter__(self):
        '''Create a generator object to iterate over the parameter space

        The parameters must be deep-copied to prevent an instance from
        modifying the class parameter space.

        :return: generator object to iterate over the parameter space.
        '''
        yield from self.__param_combinations

    def __len__(self):
        '''Returns the number of all possible parameter combinations.

        Method to calculate the test's parameter space length (i.e. the number
        of all possible parameter combinations). If the RegressionTest
        has no parameters, the length is 1.

        .. note::
           If the test is an abstract test (i.e. has undefined parameters in
           the parameter space), the returned parameter space length is 0.

        :return: length of the parameter space

        '''
        if not self.params:
            return 1

        return len(self.__param_combinations)

    def __getitem__(self, key):
        '''Access an element in the parameter space.

        If the key is an integer, this will be interpreted as a point in the
        parameter space and this function will return a mapping of the
        parameter names and their corresponding values. If the key is a
        parameter name, it will instead return all the values assigned to that
        parameter.

        If the key is an integer, this function will raise an
        :class:`IndexError` if the key is out of bounds.

        '''
        if isinstance(key, int):
            ret = {}
            val = self.__param_combinations[key]
            for i, name in enumerate(self.params):
                ret[name] = val[i]

            return ret

        try:
            return self.params[key].values
        except KeyError:
            return ()

    def is_empty(self):
        return self.params == {}

    def get_variant_nums(self, **conditions):
        '''Filter the paramter indices with a given set of conditions.

        The conditions are passed as key-value pairs, where the keys are the
        parameter names to apply the filtering on and the values are functions
        that expect the parameter's value as the sole argument.

        :returns: the indices of the matching parameters in the parameter
            space.

        '''
        candidates = range(len(self))
        if not conditions:
            return list(candidates)

        # Validate conditions
        for param, cond in conditions.items():
            if param not in self:
                raise NameError(
                    f'no such parameter: {param!r}'
                )
            elif not callable(cond):
                # Convert it to the identity function
                val = cond

                def cond(x):
                    return x == val
            elif not utils.is_trivially_callable(cond, non_def_args=1):
                raise ValueError(
                    f'condition on {param!r} must be a callable accepting a '
                    f'single argument'
                )

            def param_val(variant):
                return self._get_param_value(param, variant)

            # Filter the given condition
            candidates = [v for v in candidates if cond(param_val(v))]

        return candidates

    def _get_param_value(self, name, variant):
        '''Get the a parameter's value for a given variant.

        In this context, a variant is a point in the parameter space.
        The name argument is simply the parameter name
        '''
        return self.__param_combinations[variant][self._position[name]]



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import re

from reframe.core.exceptions import ReframeError
from reframe.core.runtime import runtime


def re_compile(patt):
    try:
        return re.compile(patt)
    except re.error:
        raise ReframeError(f'invalid regex: {patt!r}')


def _have_name(patt):
    regex = re_compile(patt)

    def _fn(case):
        # Match pattern, but remove spaces from the `display_name`
        display_name = case.check.display_name.replace(' ', '')
        rt = runtime()
        if '@' in patt:
            # Do an exact match on the unique name
            return patt.replace('@', '_') == case.check.unique_name
        elif patt.startswith('/'):
            # Do an exact match on the hashcode
            return patt[1:] == case.check.hashcode
        else:
            return regex.match(display_name)

    return _fn


def have_not_name(patt):
    def _fn(case):
        return not _have_name(patt)(case)

    return _fn


def have_any_name(names):
    rt = runtime()
    variant_matches = []
    hash_matches = []
    regex_matches = []
    for n in names:
        if '@' in n:
            test, _, variant = n.rpartition('@')
            if variant.isdigit():
                variant_matches.append((test, int(variant)))
        elif n.startswith('/'):
            hash_matches.append(n[1:])
        else:
            regex_matches.append(n)

    if regex_matches:
        regex = re_compile('|'.join(regex_matches))
    else:
        regex = None

    def _fn(case):
        # Check the variant matches
        for m in variant_matches:
            cls_name = type(case.check).__name__
            if (cls_name, case.check.variant_num) == m:
                return True

        # Check hash matches
        for m in hash_matches:
            if m == case.check.hashcode:
                return True

        display_name = case.check.display_name.replace(' ', '')
        if regex:
            return regex.match(display_name)

        return False

    return _fn


def have_tag(patt):
    regex = re_compile(patt)

    def _fn(case):
        return any(regex.match(p) for p in case.check.tags)

    return _fn


def have_not_tag(patt):
    def _fn(case):
        return not have_tag(patt)(case)

    return _fn


def have_maintainer(patt):
    regex = re_compile(patt)

    def _fn(case):
        return any(regex.match(p) for p in case.check.maintainers)

    return _fn


def have_gpu_only():
    def _fn(case):
        # NOTE: This takes into account num_gpus_per_node being None
        return case.check.num_gpus_per_node

    return _fn


def have_cpu_only():
    def _fn(case):
        # NOTE: This takes into account num_gpus_per_node being None
        return not case.check.num_gpus_per_node

    return _fn



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import decimal
import json
import jsonschema
import lxml.etree as etree
import os
import re

import reframe as rfm
import reframe.core.exceptions as errors
import reframe.utility.jsonext as jsonext
from reframe.core.warnings import suppress_deprecations

# The schema data version
# Major version bumps are expected to break the validation of previous schemas

DATA_VERSION = '3.0'
_SCHEMA = os.path.join(rfm.INSTALL_PREFIX, 'reframe/schemas/runreport.json')


class _RunReport:
    '''A wrapper to the run report providing some additional functionality'''

    def __init__(self, report):
        self._report = report
        self._fallbacks = []    # fallback reports

        # Index all runs by test case; if a test case has run multiple times,
        # only the last time will be indexed
        self._cases_index = {}
        for run in self._report['runs']:
            for tc in run['testcases']:
                c, p, e = tc['unique_name'], tc['system'], tc['environment']
                self._cases_index[c, p, e] = tc

        # Index also the restored cases
        for tc in self._report['restored_cases']:
            c, p, e = tc['unique_name'], tc['system'], tc['environment']
            self._cases_index[c, p, e] = tc

    def __getitem__(self, key):
        return self._report[key]

    def __getattr__(self, name):
        with suppress_deprecations():
            return getattr(self._report, name)

    def add_fallback(self, report):
        self._fallbacks.append(report)

    def slice(self, prop, when=None, unique=False):
        '''Slice the report on property ``prop``.'''

        if unique:
            returned = set()

        for tc in self._report['runs'][-1]['testcases']:
            val = tc[prop]
            if unique and val in returned:
                continue

            if when is None:
                if unique:
                    returned.add(val)

                yield val
            elif tc[when[0]] == when[1]:
                if unique:
                    returned.add(val)

                yield val

    def case(self, check, part, env):
        c, p, e = check.unique_name, part.fullname, env.name
        ret = self._cases_index.get((c, p, e))
        if ret is None:
            # Look up the case in the fallback reports
            for rpt in self._fallbacks:
                ret = rpt._cases_index.get((c, p, e))
                if ret is not None:
                    break

        return ret

    def restore_dangling(self, graph):
        '''Restore dangling dependencies in graph from the report data.

        Returns the updated graph.
        '''

        restored = []
        for tc, deps in graph.items():
            for d in deps:
                if d not in graph:
                    restored.append(d)
                    self._do_restore(d)

        return graph, restored

    def _do_restore(self, testcase):
        tc = self.case(*testcase)
        if tc is None:
            raise errors.ReframeError(
                f'could not restore testcase {testcase!r}: '
                f'not found in the report files'
            )

        dump_file = os.path.join(tc['stagedir'], '.rfm_testcase.json')
        try:
            with open(dump_file) as fp:
                testcase._check = jsonext.load(fp)
        except (OSError, json.JSONDecodeError) as e:
            raise errors.ReframeError(
                f'could not restore testcase {testcase!r}') from e


def next_report_filename(filepatt, new=True):
    if '{sessionid}' not in filepatt:
        return filepatt

    search_patt = os.path.basename(filepatt).replace('{sessionid}', r'(\d+)')
    new_id = -1
    basedir = os.path.dirname(filepatt) or '.'
    for filename in os.listdir(basedir):
        match = re.match(search_patt, filename)
        if match:
            found_id = int(match.group(1))
            new_id = max(found_id, new_id)

    if new:
        new_id += 1

    return filepatt.format(sessionid=new_id)


def _load_report(filename):
    try:
        with open(filename) as fp:
            report = json.load(fp)
    except OSError as e:
        raise errors.ReframeError(
            f'failed to load report file {filename!r}') from e
    except json.JSONDecodeError as e:
        raise errors.ReframeError(
            f'report file {filename!r} is not a valid JSON file') from e

    # Validate the report
    with open(_SCHEMA) as fp:
        schema = json.load(fp)

    try:
        jsonschema.validate(report, schema)
    except jsonschema.ValidationError as e:
        try:
            found_ver = report['session_info']['data_version']
        except KeyError:
            found_ver = 'n/a'

        raise errors.ReframeError(
            f'invalid report {filename!r} '
            f'(required data version: {DATA_VERSION}), found: {found_ver})'
        ) from e

    return _RunReport(report)


def load_report(*filenames):
    primary = filenames[0]
    rpt = _load_report(primary)

    # Add fallback reports
    for f in filenames[1:]:
        rpt.add_fallback(_load_report(f))

    return rpt


def junit_xml_report(json_report):
    '''Generate a JUnit report from a standard ReFrame JSON report.'''

    xml_testsuites = etree.Element('testsuites')
    for run_id, rfm_run in enumerate(json_report['runs']):
        xml_testsuite = etree.SubElement(
            xml_testsuites, 'testsuite',
            attrib={
                'errors': '0',
                'failures': str(rfm_run['num_failures']),
                'hostname': json_report['session_info']['hostname'],
                'id': str(run_id),
                'name': f'ReFrame run {run_id}',
                'package': 'reframe',
                'tests': str(rfm_run['num_cases']),
                'time': str(json_report['session_info']['time_elapsed']),

                # XSD schema does not like the timezone format, so we remove it
                'timestamp': json_report['session_info']['time_start'][:-5],
            }
        )
        testsuite_properties = etree.SubElement(xml_testsuite, 'properties')
        for tc in rfm_run['testcases']:
            casename = (
                f"{tc['unique_name']}[{tc['system']}, {tc['environment']}]"
            )
            testcase = etree.SubElement(
                xml_testsuite, 'testcase',
                attrib={
                    'classname': tc['filename'],
                    'name': casename,

                    # XSD schema does not like the exponential format and since
                    # we do not want to impose a fixed width, we pass it to
                    # `Decimal` to format it automatically.
                    'time': str(decimal.Decimal(tc['time_total'] or 0)),
                }
            )
            if tc['result'] == 'failure':
                testcase_msg = etree.SubElement(
                    testcase, 'failure', attrib={'type': 'failure',
                                                 'message': tc['fail_phase']}
                )
                testcase_msg.text = f"{tc['fail_phase']}: {tc['fail_reason']}"

        testsuite_stdout = etree.SubElement(xml_testsuite, 'system-out')
        testsuite_stdout.text = ''
        testsuite_stderr = etree.SubElement(xml_testsuite, 'system-err')
        testsuite_stderr.text = ''

    return xml_testsuites


def junit_dump(xml, fp):
    fp.write(
        etree.tostring(xml, encoding='utf8', pretty_print=True,
                       method='xml', xml_declaration=True).decode()
    )



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import abc
import logging
import logging.handlers
import numbers
import os
import re
import requests
import shutil
import socket
import sys
import time
import urllib

import reframe.utility.color as color
import reframe.utility.jsonext as jsonext
import reframe.utility.osext as osext
from reframe.core.exceptions import ConfigError, LoggingError
from reframe.core.warnings import suppress_deprecations
from reframe.utility.profile import TimeProfiler


# Global configuration options for logging
LOG_CONFIG_OPTS = {
    'handlers.filelog.prefix': './logs/'
}


# Reframe's log levels
CRITICAL = 50
ERROR    = 40
WARNING  = 30
INFO     = 20
VERBOSE  = 19
DEBUG    = 10
DEBUG2   = 9
NOTSET   = 0


_log_level_names = {
    CRITICAL: 'critical',
    ERROR:    'error',
    WARNING:  'warning',
    INFO:     'info',
    VERBOSE:  'verbose',
    DEBUG:    'debug',
    DEBUG2:   'debug2',
    NOTSET:   'undefined'
}

_log_level_values = {
    'critical':  CRITICAL,
    'error':     ERROR,
    'warning':   WARNING,
    'info':      INFO,
    'verbose':   VERBOSE,
    'debug':     DEBUG,
    'debug2':    DEBUG2,
    'undefined': NOTSET,
    'notset':    NOTSET
}


def level_from_str(name):
    return _log_level_values[name]


def _check_level(level):
    if isinstance(level, numbers.Integral):
        ret = level
    elif isinstance(level, str):
        norm_level = level.lower()
        if norm_level not in _log_level_values:
            raise ValueError(f'logger level {level} not available')
        else:
            ret = _log_level_values[norm_level]
    else:
        raise TypeError(f'logger level {level} not an int or a valid string')

    return ret


# Here we want that all the handlers of Python's logging framework understand
# our log levels. For this reason we need to do two things:
#
# 1. Monkey-patch the `setLevel` method of `logging.Handler` with our method
#    that understands our levels.
# 2. We need a way to differentiate the patched handlers. For this reason, we
#    make the `logging.Handler` a pseudo-subclass of our custom `Handler`
#    class, which itself should be abstract and unable to be instantiated.

class Handler(abc.ABC):
    @abc.abstractmethod
    def __init__(self):
        pass


Handler.register(logging.Handler)


def set_handler_level(hdlr, level):
    hdlr.level = _check_level(level)


logging.Handler.setLevel = set_handler_level


# Here we monkeypatch the `handleError` method of `logging.Handler` in
# order to ignore `BrokenPipeError` exceptions while keeping the default
# behavior for all the other types of exceptions.

def handleError(func):
    def ignore_brokenpipe(hdlr, l):
        exc_type, *_ = sys.exc_info()
        if exc_type == BrokenPipeError:
            pass
        else:
            func(hdlr, l)

    return ignore_brokenpipe


logging.Handler.handleError = handleError(logging.Handler.handleError)


class MultiFileHandler(logging.FileHandler):
    '''A file handler that allows writing on different log files based on
    information from the log record.
    '''

    def __init__(self, prefix, mode='a', encoding=None, fmt=None, perffmt=None):
        super().__init__(prefix, mode, encoding, delay=True)

        # Reset FileHandler's filename
        self.baseFilename = None
        self._prefix = prefix

        # Associates filenames with open streams
        self.__streams = {}

        # Format specifiers
        self.__fmt = fmt
        self.__perffmt = perffmt
        self.__attr_patt = re.compile(r'\%\((.*?)\)s(.)?')

    def __generate_header(self, record):
        # Generate the header from the record and fmt
        header = ''
        for m in self.__attr_patt.finditer(self.__fmt):
            attr = m.group(1)
            delim = m.group(2)
            if attr == 'check_perfvalues':
                # Expand based on the perffmt
                delim_p = ''
                for var in record.check_perfvalues:
                    name = var.split(':')[-1]
                    for m0 in self.__attr_patt.finditer(self.__perffmt):
                        attr_p  = m0.group(1)
                        delim_p = m0.group(2)
                        header += attr_p.replace('check_perf', name)
                        if delim_p:
                            header += delim_p

                if delim_p:
                    header = header.rstrip(delim_p)
            elif attr.startswith('check_'):
                header += attr[6:]
            else:
                header += attr

            if delim:
                header += delim

        return header

    def _emit_header(self, record):
        if self.baseFilename in self.__streams:
            return

        record_header = self.__generate_header(record)

        # We are opening a file for the first time;
        # check if the header has changed
        try:
            header = None
            with open(self.baseFilename) as fp:
                header = fp.readline().strip()
        except FileNotFoundError:
            return
        else:
            if header == record_header:
                return

            # Header changed; move the old file
            hcnt = 0
            while os.path.exists(self.baseFilename + f'.h{hcnt}'):
                hcnt += 1
                continue

            os.rename(self.baseFilename, self.baseFilename + f'.h{hcnt}')
        finally:
            # Open the file for writing and write the header
            fp = open(self.baseFilename,
                      mode=self.mode, encoding=self.encoding)
            if record_header != header:
                fp.write(f'{record_header}\n')

            self.__streams[self.baseFilename] = fp

    def emit(self, record):
        try:
            dirname = self._prefix % record.__dict__
            os.makedirs(dirname, exist_ok=True)
        except KeyError as e:
            raise LoggingError(f'logging failed: unknown placeholder in '
                               f'filename pattern: {e}') from None
        except OSError as e:
            raise LoggingError('logging failed') from e

        self.baseFilename = os.path.join(
            dirname, f'{record.__rfm_check__.short_name}.log'
        )
        self._emit_header(record)
        self.stream = self.__streams[self.baseFilename]
        super().emit(record)

    def close(self):
        # Close all open streams
        for s in self.__streams.values():
            self.stream = s
            super().close()


def _format_time_rfc3339(timestamp, datefmt):
    tz_suffix = time.strftime('%z', timestamp)
    tz_rfc3339 = tz_suffix[:-2] + ':' + tz_suffix[-2:]

    # Python < 3.7 truncates the `%`, whereas later versions don't
    return re.sub(r'(%)?\:z', tz_rfc3339, time.strftime(datefmt, timestamp))


def _xfmt(val):
    def _dofmt(v):
        if isinstance(v, str):
            return v
        else:
            return jsonext.dumps(v)

    # NOTE: This is for compatibility with older formatting
    if isinstance(val, (list, tuple, set)):
        return ','.join(_dofmt(v) for v in val)

    return _dofmt(val)


class CheckFieldFormatter(logging.Formatter):
    '''Log formatter that dynamically looks up format specifiers inside a
    regression test.'''

    # NOTE: This formatter will work only for the '%' style
    def __init__(self, fmt=None, datefmt=None, perffmt=None, style='%'):
        super().__init__(fmt, datefmt, style)

        self.__fmt = fmt
        self.__fmtperf = perffmt[:-1] if perffmt else ''
        self.__specs = re.findall(r'\%\((\S+?)\)s', fmt)
        self.__delim = perffmt[-1] if perffmt else ''

    def _format_perf(self, perfvars):
        chunks = []
        for var, info in perfvars.items():
            val, ref, lower, upper, unit = info
            record = {
                'check_perf_var': var.split(':')[-1],
                'check_perf_value': val,
                'check_perf_unit': unit,
                'check_perf_ref': ref,
                'check_perf_lower': lower,
                'check_perf_upper': upper
            }
            try:
                chunks.append(self.__fmtperf % record)
            except ValueError:
                chunks.append("<error formatting the performance record: "
                              "please check the 'format_perfvars' string>")

        return self.__delim.join(chunks)

    def formatMessage(self, record):
        for s in self.__specs:
            if not hasattr(record, s):
                setattr(record, s, None)

        record_proxy = dict(record.__dict__)
        for k, v in record_proxy.items():
            if k == 'check_perfvalues':
                record_proxy[k] = self._format_perf(v)
            elif k.startswith('check_'):
                record_proxy[k] = _xfmt(v)

        # Now format `check_job_completion_time` according to `datefmt`
        datefmt = self.datefmt or self.default_time_format
        ct = record.check_job_completion_time_unix
        if ct is not None:
            record_proxy['check_job_completion_time'] = _format_time_rfc3339(
                time.localtime(ct), datefmt
            )

        try:
            return self.__fmt % record_proxy
        except ValueError:
            return ("<error formatting the log message: "
                    "please check the 'format' string>")


class RFC3339Formatter(CheckFieldFormatter):
    def formatTime(self, record, datefmt=None):
        datefmt = datefmt or self.default_time_format
        if '%:z' not in datefmt:
            return super().formatTime(record, datefmt)
        else:
            timestamp = self.converter(record.created)
            return _format_time_rfc3339(timestamp, datefmt)


def _create_logger(site_config, *handlers_groups):
    level = site_config.get('logging/0/level')
    logger = Logger('reframe')
    logger.setLevel(_log_level_values[level])

    def stream_handler_kind(handler):
        if not isinstance(handler, logging.StreamHandler):
            return None
        elif handler.stream is sys.stdout:
            return 'stdout'
        elif handler.stream is sys.stderr:
            return 'stderr'
        else:
            return None

    stream_kinds = []
    for hgrp in handlers_groups:
        for handler in _extract_handlers(site_config, hgrp):
            kind = stream_handler_kind(handler)
            if kind in stream_kinds:
                continue

            if kind:
                stream_kinds.append(kind)

            logger.addHandler(handler)

    return logger


def _create_file_handler(site_config, config_prefix):
    filename = os.path.expandvars(site_config.get(f'{config_prefix}/name'))
    if not filename:
        filename = osext.mkstemp_path(suffix='.log', prefix='rfm-')

    timestamp = site_config.get(f'{config_prefix}/timestamp')
    if timestamp:
        basename, ext = os.path.splitext(filename)
        filename = f'{basename}_{time.strftime(timestamp)}{ext}'

    append = site_config.get(f'{config_prefix}/append')
    return logging.handlers.RotatingFileHandler(filename,
                                                mode='a+' if append else 'w+')


def _create_filelog_handler(site_config, config_prefix):
    basedir = os.path.abspath(os.path.join(
        site_config.get(f'systems/0/prefix'),
        osext.expandvars(site_config.get(f'{config_prefix}/basedir'))
    ))
    prefix  = osext.expandvars(site_config.get(f'{config_prefix}/prefix'))
    filename_patt = os.path.join(basedir, prefix)
    append = site_config.get(f'{config_prefix}/append')
    format = site_config.get(f'{config_prefix}/format')
    format_perf = site_config.get(f'{config_prefix}/format_perfvars')
    return MultiFileHandler(filename_patt, mode='a+' if append else 'w+',
                            fmt=format, perffmt=format_perf)


def _create_syslog_handler(site_config, config_prefix):
    address = site_config.get(f'{config_prefix}/address')

    # Check if address is in `host:port` format
    try:
        host, port = address.split(':', maxsplit=1)
    except ValueError:
        pass
    else:
        try:
            address = (host, int(port))
        except ValueError:
            raise ConfigError(
                f'syslog address port not an integer: {port!r}'
            ) from None

    facility = site_config.get(f'{config_prefix}/facility')
    try:
        facility_type = logging.handlers.SysLogHandler.facility_names[facility]
    except KeyError:
        # This should not happen
        raise AssertionError(
            f'syslog handler: unknown facility: {facility}') from None

    socktype = site_config.get(f'{config_prefix}/socktype')
    if socktype == 'udp':
        socket_type = socket.SOCK_DGRAM
    elif socktype == 'tcp':
        socket_type = socket.SOCK_STREAM
    else:
        # This should not happen
        raise AssertionError(
            f'syslog handler: unknown socket type: {socktype}'
        )

    return logging.handlers.SysLogHandler(address, facility_type, socket_type)


def _create_stream_handler(site_config, config_prefix):
    stream = site_config.get(f'{config_prefix}/name')
    if stream == 'stdout':
        return logging.StreamHandler(stream=sys.stdout)
    elif stream == 'stderr':
        return logging.StreamHandler(stream=sys.stderr)
    else:
        # This should not happen
        raise AssertionError(f'unknown stream: {stream}')


def _create_graylog_handler(site_config, config_prefix):
    try:
        import pygelf
    except ImportError as err:
        getlogger().warning(f'could not import pygelf module: {err}')
        return None

    address = site_config.get(f'{config_prefix}/address')
    host, *port = address.split(':', maxsplit=1)
    if not port:
        raise ConfigError('graylog handler: no port specified')

    port = port[0]

    # Check if the remote server is up and accepts connections; if not we will
    # skip the handler
    try:
        with socket.create_connection((host, port), timeout=1):
            pass
    except OSError as e:
        getlogger().warning(
            f"could not connect to Graylog server at '{address}': {e}"
        )
        return None

    extras = site_config.get(f'{config_prefix}/extras')
    return pygelf.GelfHttpHandler(host=host, port=port,
                                  static_fields=extras,
                                  include_extra_fields=True,
                                  json_default=jsonext.encode)


def _create_httpjson_handler(site_config, config_prefix):
    url = site_config.get(f'{config_prefix}/url')
    parsed_url = urllib.parse.urlparse(url)
    if parsed_url.scheme not in {'http', 'https'}:
        raise ConfigError(
            "httpjson handler: invalid url scheme: use 'http' or 'https'"
        )

    if not parsed_url.hostname:
        raise ConfigError('http json handler: invalid hostname')

    try:
        if not parsed_url.port:
            raise ConfigError('http json handler: no port given')
    except ValueError as e:
        raise ConfigError('http json handler: invalid port') from e

    # Check if the remote server is up and accepts connections; if not we will
    # skip the handler
    try:
        with socket.create_connection((parsed_url.hostname, parsed_url.port),
                                      timeout=1):
            pass
    except OSError as e:
        getlogger().warning(
            f'httpjson: could not connect to server '
            f'{parsed_url.hostname}:{parsed_url.port}: {e}'
        )
        return None

    extras = site_config.get(f'{config_prefix}/extras')
    ignore_keys = site_config.get(f'{config_prefix}/ignore_keys')
    return HTTPJSONHandler(url, extras, ignore_keys)


class HTTPJSONHandler(logging.Handler):
    # These are the `LogRecord` attributes which are defined in
    # https://docs.python.org/3/library/logging.html#logrecord-attributes
    # as well as the 'exc_text' which is used to cache the traceback text
    LOG_ATTRS = {
        'args', 'asctime', 'created', 'exc_info',
        'filename', 'funcName', 'levelname', 'levelno',
        'lineno', 'message', 'module', 'msecs', 'msg', 'name',
        'pathname', 'process', 'processName', 'relativeCreated',
        'stack_info', 'thread', 'threadName', 'exc_text'
    }

    def __init__(self, url, extras=None, ignore_keys=None):
        super().__init__()
        self._url = url
        self._extras = extras
        self._ignore_keys = ignore_keys

    def _record_to_json(self, record):
        def _can_send(key):
            return not (
                key.startswith('_') or key in HTTPJSONHandler.LOG_ATTRS or
                (self._ignore_keys and key in self._ignore_keys)
            )

        json_record = {
            k: v for k, v in record.__dict__.items() if _can_send(k)
        }
        if self._extras:
            json_record.update({
                k: v for k, v in self._extras.items() if _can_send(k)
            })

        return _xfmt(json_record).encode('utf-8')

    def emit(self, record):
        json_record = self._record_to_json(record)
        try:
            requests.post(
                self._url, data=json_record,
                headers={'Content-type': 'application/json'}
            )
        except requests.exceptions.RequestException as e:
            raise LoggingError('logging failed') from e


def _extract_handlers(site_config, handlers_group):
    handler_prefix = f'logging/0/{handlers_group}'
    handlers_list = site_config.get(handler_prefix)
    handlers = []
    for i, handler_config in enumerate(handlers_list):
        handler_type = handler_config['type']
        if handler_type == 'file':
            hdlr = _create_file_handler(site_config, f'{handler_prefix}/{i}')
        elif handler_type == 'filelog':
            hdlr = _create_filelog_handler(
                site_config, f'{handler_prefix}/{i}'
            )
        elif handler_type == 'syslog':
            hdlr = _create_syslog_handler(site_config, f'{handler_prefix}/{i}')
        elif handler_type == 'stream':
            hdlr = _create_stream_handler(site_config, f'{handler_prefix}/{i}')
        elif handler_type == 'graylog':
            hdlr = _create_graylog_handler(
                site_config, f'{handler_prefix}/{i}'
            )
            if hdlr is None:
                getlogger().warning('could not initialize the '
                                    'graylog handler; ignoring ...')
                continue
        elif handler_type == 'httpjson':
            hdlr = _create_httpjson_handler(
                site_config, f'{handler_prefix}/{i}'
            )
            if hdlr is None:
                getlogger().warning('could not initialize the '
                                    'httpjson handler; ignoring ...')
                continue
        else:
            # Should not enter here
            raise AssertionError(f'unknown handler type: {handler_type}')

        level = site_config.get(f'{handler_prefix}/{i}/level')
        fmt = site_config.get(f'{handler_prefix}/{i}/format')
        datefmt = site_config.get(f'{handler_prefix}/{i}/datefmt')
        perffmt = site_config.get(f'{handler_prefix}/{i}/format_perfvars')
        hdlr.setFormatter(RFC3339Formatter(fmt=fmt,
                                           datefmt=datefmt, perffmt=perffmt))
        hdlr.setLevel(_check_level(level))
        handlers.append(hdlr)

    return handlers


class Logger(logging.Logger):
    def __init__(self, name, level=logging.NOTSET):
        # We will set the logger level ourselves so as to bypass the base
        # class' check
        super().__init__(name, logging.NOTSET)
        self.level = _check_level(level)

    def setLevel(self, level):
        self.level = _check_level(level)

        if sys.version_info[:2] >= (3, 7):
            # Clear the internal cache of the base logger, otherwise the
            # logger will remain disabled if its level is raised and then
            # lowered again
            self._cache.clear()

    def makeRecord(self, name, level, fn, lno, msg, args, exc_info,
                   func=None, extra=None, sinfo=None):
        record = super().makeRecord(name, level, fn, lno, msg, args, exc_info,
                                    func, extra, sinfo)
        try:
            # Fill in our name for the record
            record.levelname = _log_level_names[level]
        except KeyError:
            # Go with the default level name of Python logging
            pass

        return record

    # Override all the convenience logging functions, because we want to make
    # sure that they map to our level definitions

    def critical(self, msg, *args, **kwargs):
        return self.log(CRITICAL, msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        return self.log(ERROR, msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        self.log(WARNING, msg, *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        self.log(INFO, msg, *args, **kwargs)

    def verbose(self, message, *args, **kwargs):
        self.log(VERBOSE, message, *args, **kwargs)

    def debug(self, message, *args, **kwargs):
        self.log(DEBUG, message, *args, **kwargs)

    def debug2(self, message, *args, **kwargs):
        self.log(DEBUG2, message, *args, **kwargs)


# This is a cache for warnings that we don't want to repeat
_WARN_ONCE = set()


class LoggerAdapter(logging.LoggerAdapter):
    def __init__(self, logger=None, check=None):
        super().__init__(
            logger,
            {
                # Here we only set the format specifiers that do not
                # correspond directly to check attributes
                '__rfm_check__': check,
                'check_name': 'reframe',
                'check_jobid': None,
                'check_job_completion_time': None,
                'check_job_completion_time_unix': None,
                'check_info': 'reframe',
                'check_system': None,
                'check_partition': None,
                'check_environ': None,
                'check_perf_var': None,
                'check_perf_value': None,
                'check_perf_ref': None,
                'check_perf_lower_thres': None,
                'check_perf_upper_thres': None,
                'check_perf_unit': None,
                'check_result': None,
                'osuser':  osext.osuser(),
                'osgroup': osext.osgroup(),
                'version': osext.reframe_version(),
            }
        )
        self.check = check
        self.colorize = False

    def setLevel(self, level):
        if self.logger:
            super().setLevel(level)

    @property
    def std_stream_handlers(self):
        if self.logger:
            return [h for h in self.logger.handlers
                    if isinstance(h, logging.StreamHandler)]
        else:
            return []

    def _update_check_extras(self):
        '''Return a dictionary with all the check-specific information.'''

        if self.check is None:
            return

        check_type = type(self.check)
        for attr, alt_name in check_type.loggable_attrs():
            extra_name  = alt_name or attr

            with suppress_deprecations():
                # In case of AttributeError, i.e., the variable is undefined,
                # we set the value to None
                val = getattr(self.check, attr, None)

            if attr in check_type.raw_params:
                # Attribute is parameter, so format it
                val = check_type.raw_params[attr].format(val)

            self.extra[f'check_{extra_name}'] = val

        # Add special extras
        self.extra['check_info'] = self.check.info()
        self.extra['check_job_completion_time'] = _format_time_rfc3339(
            time.localtime(self.extra['check_job_completion_time_unix']),
            '%FT%T%:z'
        )

    def log_performance(self, level, task, msg=None, multiline=False):
        if self.check is None or not self.check.is_performance_check():
            return

        self.extra['check_partition'] = task.testcase.partition.name
        self.extra['check_environ'] = task.testcase.environ.name
        self.extra['check_result'] = 'pass' if task.succeeded else 'fail'
        if msg is None:
            msg = 'sent by ' + self.extra['osuser']

        if multiline:
            # Log one record for each performance variable
            for var, info in self.check.perfvalues.items():
                val, ref, lower, upper, unit = info
                self.extra['check_perf_var'] = var.split(':')[-1]
                self.extra['check_perf_value'] = val
                self.extra['check_perf_ref'] = ref
                self.extra['check_perf_lower_thres'] = lower
                self.extra['check_perf_upper_thres'] = upper
                self.extra['check_perf_unit'] = unit
                self.log(level, msg)
        else:
            self.log(level, msg)

    def process(self, msg, kwargs):
        # Setup dynamic fields of the check
        self._update_check_extras()
        try:
            self.extra.update(kwargs['extra'])
        except KeyError:
            pass

        return super().process(msg, kwargs)

    # Override log() function to treat `None` loggers
    def log(self, level, msg, *args, **kwargs):
        if self.logger:
            super().log(level, msg, *args, **kwargs)

    def debug2(self, message, *args, **kwargs):
        self.log(DEBUG2, message, *args, **kwargs)

    def verbose(self, message, *args, **kwargs):
        self.log(VERBOSE, message, *args, **kwargs)

    def warning(self, message, *args, cache=False, **kwargs):
        if cache:
            if message in _WARN_ONCE:
                return

            _WARN_ONCE.add(message)

        message = f'WARNING: {message}'
        if self.colorize:
            message = color.colorize(message, color.YELLOW)

        super().warning(message, *args, **kwargs)

    def error(self, message, *args, **kwargs):
        message = f'ERROR: {message}'
        if self.colorize:
            message = color.colorize(message, color.RED)

        super().error(message, *args, **kwargs)

    def adjust_verbosity(self, num_steps):
        '''Convenience function for increasing or decreasing the verbosity
        of the logger step-wise.'''
        log_levels = sorted(_log_level_names.keys())[1:]
        for h in self.std_stream_handlers:
            level_idx = log_levels.index(h.level)
            new_level_idx = level_idx - num_steps
            if new_level_idx < 0:
                new_level = log_levels[0]
            elif new_level_idx >= len(log_levels):
                new_level = log_levels[-1]
            else:
                new_level = log_levels[new_level_idx]

            h.setLevel(new_level)


# A logger that doesn't log anything
null_logger = LoggerAdapter()

_logger = None
_perf_logger = None
_context_logger = null_logger


class logging_context:
    def __init__(self, check=None, level=DEBUG):
        global _context_logger

        self._level = level
        self._orig_logger = _context_logger
        if check is not None:
            _context_logger = LoggerAdapter(_logger, check)
            _context_logger.colorize = self._orig_logger.colorize

    def __enter__(self):
        return _context_logger

    def __exit__(self, exc_type, exc_value, traceback):
        global _context_logger

        # Log any exceptions thrown with the current context logger
        if exc_type is not None:
            msg = 'caught {0}: {1}'
            exc_fullname = f'{exc_type.__module__}.{exc_type.__name__}'
            getlogger().log(self._level, msg.format(exc_fullname, exc_value))

        # Restore context logger
        _context_logger = self._orig_logger


def configure_logging(site_config):
    global _logger, _context_logger, _perf_logger

    if site_config is None:
        _logger = None
        _context_logger = null_logger
        return

    _logger = _create_logger(site_config, 'handlers$', 'handlers')
    _perf_logger = _create_logger(site_config, 'handlers_perflog')
    _context_logger = LoggerAdapter(_logger)


def log_files():
    return [hdlr.baseFilename for hdlr in _logger.handlers
            if isinstance(hdlr, logging.FileHandler)]


def save_log_files(dest):
    os.makedirs(dest, exist_ok=True)
    return [shutil.copy(logfile, dest, follow_symlinks=True)
            for logfile in log_files()]


def getlogger():
    return _context_logger


def getperflogger(check):
    return LoggerAdapter(_perf_logger, check)


# Global framework profiler
_profiler = TimeProfiler()


def getprofiler():
    return _profiler


def time_function(fn):
    '''Decorator for timing a function using the global profiler'''

    def _fn(*args, **kwargs):
        with _profiler.time_region(fn.__qualname__):
            return fn(*args, **kwargs)

    return _fn


def time_function_noexit(fn):
    '''Decorator for timing a function using the global profiler'''

    def _fn(*args, **kwargs):
        _profiler.enter_region(fn.__qualname__)
        return fn(*args, **kwargs)

    return _fn



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

# A lightweight time profiler

import time
import sys

from collections import OrderedDict


class ProfilerError(Exception):
    pass


class time_region:
    '''Context manager for timing a code region'''

    def __init__(self, region, profiler=None):
        self._profiler = profiler or TimeProfiler()
        self._region = region

    def __enter__(self):
        self._profiler.enter_region(self._region)
        return self._profiler

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._profiler.exit_region()


class TimeProfiler:
    def __init__(self):
        self._region_stack = ['root']
        if sys.version_info[:2] < (3, 8):
            self._region_times = OrderedDict()
        else:
            self._region_times = {}

    @property
    def current_region(self):
        return self._region_stack[-1]

    def enter_region(self, region_name):
        timestamp = time.time()
        region_fullname = f'{self.current_region}:{region_name}'
        if region_fullname in self._region_times:
            elapsed = self._region_times[region_fullname][1]
        else:
            elapsed = 0.0

        self._region_times[region_fullname] = (timestamp, elapsed)
        self._region_stack.append(region_fullname)

    def exit_region(self):
        timestamp = time.time()
        region = self.current_region
        t_start, elapsed = self._region_times[region]
        self._region_times[region] = (None, elapsed + timestamp - t_start)
        self._region_stack.pop()

    def total_time(self, region_name):
        for region in reversed(self._region_times.keys()):
            if (region == region_name or
                region.rsplit(':', maxsplit=1)[-1] == region_name):
                timestamp, elapsed = self._region_times[region]
                if timestamp:
                    raise ProfilerError(
                        f'region {region_name!r} has not exited'
                    )

                return elapsed

        raise ProfilerError(f'unknown region: {region_name!r}')

    def time_region(self, region):
        return globals()['time_region'](region, self)

    def print_report(self, print_fn=None):
        if print_fn is None:
            print_fn = print

        print_fn('>>> profiler report [start] <<<')
        for name, t_info in self._region_times.items():
            # Remove the root prefix
            levels = name.count(':')
            indent = ' '*4*(levels - 1)
            region_name = name.rsplit(':', maxsplit=1)[-1]
            msg = f'{indent}{region_name}: {t_info[1]:.6f} s'
            if t_info[0]:
                msg += ' <incomplete>'

            print_fn(msg)

        print_fn('>>> profiler report [ end ] <<<')
